#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import dataclasses
from typing import List, Optional, Type

from fastapi import Depends

from argilla_server import telemetry
from argilla_server.commons.config import TasksFactory
from argilla_server.commons.models import TaskStatus
from argilla_server.daos.backend.base import WrongLogDataError
from argilla_server.daos.records import DatasetRecordsDAO
from argilla_server.errors import BulkDataError, ForbiddenOperationError
from argilla_server.models import User
from argilla_server.policies import DatasetPolicy, is_authorized
from argilla_server.services.datasets import ServiceDataset
from argilla_server.services.search.model import ServiceBaseRecordsQuery
from argilla_server.services.tasks.commons import ServiceRecord


@dataclasses.dataclass
class DeleteRecordsOut:
    processed: int = 0
    discarded: int = 0
    deleted: int = 0


class RecordsStorageService:
    _INSTANCE: "RecordsStorageService" = None

    @classmethod
    def get_instance(
        cls,
        dao: DatasetRecordsDAO = Depends(DatasetRecordsDAO.get_instance),
    ) -> "RecordsStorageService":
        if not cls._INSTANCE:
            cls._INSTANCE = cls(dao)
        return cls._INSTANCE

    def __init__(self, dao: DatasetRecordsDAO):
        self.__dao__ = dao

    async def store_records(
        self,
        dataset: ServiceDataset,
        records: List[ServiceRecord],
        record_type: Type[ServiceRecord],
    ) -> int:
        """Store a set of records"""
        await telemetry.track_bulk(task=dataset.task, records=len(records))

        metrics = TasksFactory.get_task_metrics(dataset.task)
        if metrics:
            for record in records:
                record.metrics = metrics.record_metrics(record)

        try:
            return self.__dao__.add_records(
                dataset=dataset,
                records=records,
                record_class=record_type,
            )
        except WrongLogDataError as ex:
            raise BulkDataError(
                detail=f"Cannot log data in dataset {dataset.id}",
                errors=ex.errors,
            )

    async def delete_records(
        self,
        user: User,
        dataset: ServiceDataset,
        query: Optional[ServiceBaseRecordsQuery] = None,
        mark_as_discarded: bool = False,
    ) -> DeleteRecordsOut:
        processed, discarded, deleted = None, None, None
        if mark_as_discarded:
            processed, discarded = await self.__dao__.update_records_by_query(
                dataset,
                query=query,
                status=TaskStatus.discarded,
            )
        else:
            if not await is_authorized(user, DatasetPolicy.delete_records(dataset)):
                raise ForbiddenOperationError(
                    "You don't have the necessary permissions to delete records on this dataset. "
                    "Only administrators can delete datasets"
                )

            processed, deleted = await self.__dao__.delete_records_by_query(dataset, query=query)

        return DeleteRecordsOut(
            processed=processed or 0,
            discarded=discarded or 0,
            deleted=deleted or 0,
        )

    async def update_record(
        self,
        dataset: ServiceDataset,
        record: ServiceRecord,
        **data,
    ) -> ServiceRecord:
        if data.get("metadata"):
            record.metadata = {
                **(record.metadata or {}),
                **data.pop("metadata"),
            }

        new_record = record.copy(update=data)  # Data validation
        await self.__dao__.update_record(
            dataset=dataset,
            record=record,
        )
        return new_record



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import logging
from typing import Tuple

from opensearchpy import OpenSearch
from packaging.version import parse

from argilla_server.daos.backend.base import GenericSearchError
from argilla_server.daos.backend.client_adapters import ElasticsearchClient, IClientAdapter, OpenSearchClient
from argilla_server.settings import settings as server_settings

_LOGGER = logging.getLogger("argilla")


class ClientAdapterFactory:
    @classmethod
    def get(
        cls,
        hosts: str,
        index_shards: int,
        ssl_verify: bool,
        ca_path: str,
        retry_on_timeout: bool = True,
        max_retries: int = 5,
    ) -> IClientAdapter:
        client_config = dict(
            hosts=hosts,
            verify_certs=ssl_verify,
            ca_certs=ca_path,
            retry_on_timeout=retry_on_timeout,
            max_retries=max_retries,
        )

        version, distribution = cls._fetch_cluster_version_info(client_config)

        if distribution != server_settings.search_engine:
            _LOGGER.warning(
                f"The search engine distribution is not the same as the one configured in the server. "
                f"The configuration will be automatically updated with ARGILLA_SEARCH_ENGINE={distribution}."
            )
            server_settings.search_engine = distribution

        if distribution == "elasticsearch" and parse("8.5.0") <= parse(version):
            client_class = ElasticsearchClient
        elif distribution == "opensearch" and parse("2.4.0") <= parse(version):
            client_class = OpenSearchClient
        else:
            raise ValueError(
                f"Unsupported ElasticSearch/OpenSearch backend version :{version}. "
                f"Minimal supported version are `8.5.0` for Elasticsearch and `2.4.0` for OpenSearch. "
                "Please, upgrade the backend to a supported version."
            )

        return client_class(index_shards=index_shards, config_backend=client_config)

    @classmethod
    def _fetch_cluster_version_info(cls, client_config: dict) -> Tuple[str, str]:
        try:
            # All security config will be used here.
            # See here https://opensearch.org/docs/latest/clients/index/#legacy-clients
            client = OpenSearch(**client_config)
            data = client.info()

            version_info = data["version"]
            version: str = version_info["number"]
            distribution: str = version_info.get("distribution", "elasticsearch")

            return version, distribution
        except Exception as error:
            raise GenericSearchError(error)



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pytest
from argilla_server.apis.v0.models.token_classification import (
    TokenClassificationAnnotation,
    TokenClassificationQuery,
    TokenClassificationRecord,
)
from argilla_server.commons.models import PredictionStatus
from argilla_server.daos.backend.search.query_builder import EsQueryBuilder
from argilla_server.services.tasks.token_classification.model import (
    EntitySpan,
    ServiceTokenClassificationRecord,
)

from tests.pydantic_v1 import ValidationError


def test_char_position():
    with pytest.raises(
        ValidationError,
        match="End character cannot be placed before the starting character,"
        " it must be at least one character after.",
    ):
        EntitySpan(start=1, end=1, label="label")

    text = "I am Maxi"
    ServiceTokenClassificationRecord(
        text=text,
        tokens=text.split(),
        prediction=TokenClassificationAnnotation(
            agent="test",
            entities=[
                EntitySpan(start=0, end=1, label="test"),
                EntitySpan(start=2, end=len(text), label="test"),
            ],
        ),
    )


def test_fix_substrings():
    text = "On one ones o no"
    ServiceTokenClassificationRecord(
        text=text,
        tokens=text.split(),
        prediction=TokenClassificationAnnotation(
            agent="test",
            entities=[
                EntitySpan(start=3, end=6, label="test"),
            ],
        ),
    )


def test_entities_with_spaces():
    text = "This is  a  great  space"
    ServiceTokenClassificationRecord(
        text=text,
        tokens=["This", "is", " ", "a", " ", "great", " ", "space"],
        prediction=TokenClassificationAnnotation(
            agent="test",
            entities=[
                EntitySpan(start=9, end=len(text), label="test"),
            ],
        ),
    )


def test_model_dict():
    text = "This is  a  great  space"
    tokens = ["This", "is", " ", "a", " ", "great", " ", "space"]
    record = TokenClassificationRecord(
        id="1",
        text=text,
        tokens=tokens,
        prediction=TokenClassificationAnnotation(agent="test", entities=[EntitySpan(start=9, end=24, label="test")]),
    )

    assert record.dict(exclude_none=True) == {
        "id": "1",
        "metrics": {},
        "prediction": {
            "agent": "test",
            "entities": [{"end": 24, "label": "test", "score": 1.0, "start": 9}],
        },
        "predictions": {
            "test": {
                "entities": [{"end": 24, "label": "test", "score": 1.0, "start": 9}],
            }
        },
        "status": "Default",
        "text": "This is  a  great  space",
        "tokens": ["This", "is", " ", "a", " ", "great", " ", "space"],
    }


def test_model_with_predictions():
    record = TokenClassificationRecord.parse_obj(
        {
            "id": 1,
            "metrics": {},
            "predictions": {
                "test": {
                    "entities": [{"end": 24, "label": "test", "score": 1.0, "start": 9}],
                }
            },
            "status": "Default",
            "text": "This is  a  great  space",
            "tokens": ["This", "is", " ", "a", " ", "great", " ", "space"],
        }
    )
    assert record.dict(exclude_none=True) == {
        "id": 1,
        "metrics": {},
        "prediction": {
            "agent": "test",
            "entities": [{"end": 24, "label": "test", "score": 1.0, "start": 9}],
        },
        "predictions": {
            "test": {
                "entities": [{"end": 24, "label": "test", "score": 1.0, "start": 9}],
            }
        },
        "status": "Default",
        "text": "This is  a  great  space",
        "tokens": ["This", "is", " ", "a", " ", "great", " ", "space"],
    }


def test_too_long_metadata():
    text = "On one ones o no"
    record = ServiceTokenClassificationRecord.parse_obj(
        {
            "text": text,
            "tokens": text.split(),
            "metadata": {"too_long": "a" * 1000},
        }
    )

    assert len(record.metadata["too_long"]) == 128


def test_entity_label_too_long():
    text = "On one ones o no"
    with pytest.raises(ValidationError, match="ensure this value has at most 128 character"):
        ServiceTokenClassificationRecord(
            text=text,
            tokens=text.split(),
            prediction=TokenClassificationAnnotation(
                agent="test",
                entities=[
                    EntitySpan(
                        start=9,
                        end=len(text),
                        label="a" * 1000,
                    ),
                ],
            ),
        )


def test_to_es_query():
    query = TokenClassificationQuery(ids=[1, 2, 3])
    assert EsQueryBuilder._to_es_query(query) == {"ids": {"values": query.ids}}


def test_misaligned_entity_mentions_with_spaces_left():
    assert ServiceTokenClassificationRecord(
        text="according to analysts.\n     Dart Group Corp was not",
        tokens=[
            "according",
            "to",
            "analysts",
            ".",
            "\n     ",
            "Dart",
            "Group",
            "Corp",
            "was",
            "not",
        ],
        annotation=TokenClassificationAnnotation(
            agent="heuristics",
            entities=[EntitySpan(start=22, end=43, label="COMPANY", score=1.0)],
            score=None,
        ),
    )


def test_misaligned_entity_mentions_with_spaces_right():
    assert ServiceTokenClassificationRecord(
        text="\nvs 9.91 billion\n    Note\n REUTER\n",
        tokens=["\n", "vs", "9.91", "billion", "\n    ", "Note", "\n ", "REUTER", "\n"],
        annotation=TokenClassificationAnnotation(
            agent="heuristics",
            entities=[EntitySpan(start=4, end=21, label="MONEY", score=1.0)],
            score=None,
        ),
    )


def test_custom_tokens_splitting():
    ServiceTokenClassificationRecord(
        text="ThisisMr.Bean, a character  playedby actor RowanAtkinson",
        tokens=[
            "This",
            "is",
            "Mr.",
            "Bean",
            "a",
            "character",
            "played",
            "by",
            "actor",
            "Rowan",
            "Atkinson",
        ],
        annotation=TokenClassificationAnnotation(
            agent="test",
            entities=[
                EntitySpan(start=9, end=13, label="PERSON"),
                EntitySpan(start=43, end=48, label="NAME"),
                EntitySpan(start=48, end=56, label="SURNAME"),
            ],
        ),
    )


def test_record_scores():
    record = ServiceTokenClassificationRecord(
        text="\nvs 9.91 billion\n    Note\n REUTER\n",
        tokens=["\n", "vs", "9.91", "billion", "\n    ", "Note", "\n ", "REUTER", "\n"],
        prediction=TokenClassificationAnnotation(
            agent="heuristics",
            entities=[
                EntitySpan(start=4, end=21, label="MONEY", score=0.8),
                EntitySpan(start=4, end=21, label="ORG", score=0.1),
                EntitySpan(start=4, end=21, label="PERSON", score=0.2),
            ],
        ),
    )
    assert record.scores == [0.8, 0.1, 0.2]


def test_annotated_without_entities():
    text = "The text that i wrote"
    record = ServiceTokenClassificationRecord(
        text=text,
        tokens=text.split(),
        prediction=TokenClassificationAnnotation(agent="pred.test", entities=[EntitySpan(start=0, end=3, label="DET")]),
        annotation=TokenClassificationAnnotation(agent="test", entities=[]),
    )

    assert record.annotated_by == [record.annotation.agent]
    assert record.predicted_by == [record.prediction.agent]
    assert record.predicted == PredictionStatus.KO


def test_adjust_spans():
    text = "A text with  some empty     spaces  that could    bring  not cleany   annotated spans"
    record = ServiceTokenClassificationRecord(
        text=text,
        tokens=text.split(),
        prediction=TokenClassificationAnnotation(
            agent="pred.test",
            entities=[
                EntitySpan(start=-3, end=2, label="DET"),
                EntitySpan(start=24, end=36, label="NAME"),
            ],
        ),
        annotation=TokenClassificationAnnotation(
            agent="test",
            entities=[
                EntitySpan(start=48, end=61, label="VERB"),
                EntitySpan(start=68, end=100, label="DET"),
            ],
        ),
    )

    assert record.prediction.entities == [
        EntitySpan(start=0, end=1, label="DET"),
        EntitySpan(start=28, end=34, label="NAME"),
    ]

    assert record.annotation.entities == [
        EntitySpan(start=50, end=60, label="VERB"),
        EntitySpan(start=70, end=85, label="DET"),
    ]


def test_whitespace_in_tokens():
    data = {
        "text": "every four (4)  ",
        "tokens": ["every", "four", "(", "4", ")", " "],
        "prediction": {"agent": "mock", "entities": [{"end": 16, "label": "mock", "start": 0}]},
    }

    record = ServiceTokenClassificationRecord.parse_obj(data)
    assert record.tokens == ["every", "four", "(", "4", ")", " "]


def test_predicted_ok_ko_computation():
    text = "A text with some empty spaces that could bring not cleanly annotated spans"
    record = ServiceTokenClassificationRecord(
        text=text,
        tokens=text.split(),
        prediction=TokenClassificationAnnotation(
            agent="pred.test",
            entities=[
                EntitySpan(
                    start=0,
                    end=6,
                    label="VERB",
                ),
                EntitySpan(
                    start=47,
                    end=68,
                    label="VERB",
                    score=0.5,
                ),
            ],
        ),
        annotation=TokenClassificationAnnotation(
            agent="test",
            entities=[
                EntitySpan(
                    start=0,
                    end=6,
                    label="VERB",
                ),
                EntitySpan(
                    start=47,
                    end=68,
                    label="VERB",
                ),
            ],
        ),
    )

    assert record.predicted == PredictionStatus.OK

    record.annotation.entities = record.annotation.entities[1:]
    assert record.predicted == PredictionStatus.KO



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from datetime import datetime
from typing import Annotated, List, Literal, Optional
from uuid import UUID

from typing_extensions import Annotated

from argilla_server.enums import FieldType
from argilla_server.pydantic_v1 import BaseModel, constr
from argilla_server.pydantic_v1 import Field as PydanticField
from argilla_server.schemas.base import UpdateSchema

FIELD_CREATE_NAME_REGEX = r"^(?=.*[a-z0-9])[a-z0-9_-]+$"
FIELD_CREATE_NAME_MIN_LENGTH = 1
FIELD_CREATE_NAME_MAX_LENGTH = 200

FIELD_CREATE_TITLE_MIN_LENGTH = 1
FIELD_CREATE_TITLE_MAX_LENGTH = 500


FieldName = Annotated[
    constr(
        regex=FIELD_CREATE_NAME_REGEX,
        min_length=FIELD_CREATE_NAME_MIN_LENGTH,
        max_length=FIELD_CREATE_NAME_MAX_LENGTH,
    ),
    PydanticField(..., description="The name of the field"),
]


FieldTitle = Annotated[
    constr(
        min_length=FIELD_CREATE_TITLE_MIN_LENGTH,
        max_length=FIELD_CREATE_TITLE_MAX_LENGTH,
    ),
    PydanticField(..., description="The title of the field"),
]


class TextFieldSettings(BaseModel):
    type: Literal[FieldType.text]
    use_markdown: bool = False


class TextFieldSettingsUpdate(UpdateSchema):
    type: Literal[FieldType.text]
    use_markdown: bool


class Field(BaseModel):
    id: UUID
    name: str
    title: str
    required: bool
    settings: TextFieldSettings
    dataset_id: UUID
    inserted_at: datetime
    updated_at: datetime

    class Config:
        orm_mode = True


class Fields(BaseModel):
    items: List[Field]


class FieldCreate(BaseModel):
    name: FieldName
    title: FieldTitle
    required: Optional[bool]
    settings: TextFieldSettings


class FieldUpdate(UpdateSchema):
    title: Optional[FieldTitle]
    settings: Optional[TextFieldSettingsUpdate]

    __non_nullable_fields__ = {"title", "settings"}



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import os
from typing import Any, Dict

import psutil
from fastapi import Depends

from argilla_server._version import __version__ as version
from argilla_server.daos.backend import GenericElasticEngineBackend
from argilla_server.pydantic_v1 import BaseModel


def size(bytes):
    system = [
        (1024**5, "P"),
        (1024**4, "T"),
        (1024**3, "G"),
        (1024**2, "M"),
        (1024**1, "K"),
        (1024**0, "B"),
    ]

    factor, suffix = None, None
    for factor, suffix in system:
        if bytes >= factor:
            break

    amount = int(bytes / factor)
    if isinstance(suffix, tuple):
        singular, multiple = suffix
        if amount == 1:
            suffix = singular
        else:
            suffix = multiple
    return str(amount) + suffix


class ApiInfo(BaseModel):
    """Basic api info"""

    version: str


class ApiStatus(ApiInfo):
    """The argilla api status model"""

    elasticsearch: Dict[str, Any]
    mem_info: Dict[str, Any]


class ApiInfoService:
    """
    The api info service
    """

    _INSTANCE = None

    @classmethod
    def get_instance(
        cls,
        backend: GenericElasticEngineBackend = Depends(GenericElasticEngineBackend.get_instance),
    ) -> "ApiInfoService":
        """
        Creates an api info service
        """

        if not cls._INSTANCE:
            cls._INSTANCE = ApiInfoService(backend)
        return cls._INSTANCE

    def __init__(self, es: GenericElasticEngineBackend):
        self.__es__ = es

    def api_status(self) -> ApiStatus:
        """Returns the current api status"""
        return ApiStatus(
            version=str(version),
            elasticsearch=self._elasticsearch_info(),
            mem_info=self._api_memory_info(),
        )

    def _elasticsearch_info(self) -> Dict[str, Any]:
        """Returns the elasticsearch cluster info"""
        return self.__es__.client.get_cluster_info()

    @staticmethod
    def _api_memory_info() -> Dict[str, Any]:
        """Fetch the api process memory usage"""
        process = psutil.Process(os.getpid())
        return {k: size(v) for k, v in process.memory_info()._asdict().items()}



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from uuid import UUID, uuid4

import pytest
from argilla_server.enums import SuggestionType
from httpx import AsyncClient

from tests.factories import RecordFactory, TextQuestionFactory


@pytest.mark.asyncio
class TestUpsertSuggestion:
    def url(self, record_id: UUID) -> str:
        return f"/api/v1/records/{record_id}/suggestions"

    @pytest.mark.parametrize(
        "agent", ["a", "A", "0", "gpt 3.5", "gpt-3.5-turbo", "argilla/zephyr-7b", "ft:gpt-3.5-turbo"]
    )
    async def test_with_valid_agent(self, async_client: AsyncClient, owner_auth_header: dict, agent: str):
        record = await RecordFactory.create()
        question = await TextQuestionFactory.create(dataset=record.dataset)

        response = await async_client.put(
            self.url(record.id),
            headers=owner_auth_header,
            json={
                "question_id": str(question.id),
                "type": SuggestionType.model,
                "value": "value",
                "agent": agent,
                "score": 1.0,
            },
        )

        assert response.status_code == 201

    @pytest.mark.parametrize("agent", ["", " ", "  ", "-", "_", ":", ".", "/", ","])
    async def test_with_invalid_agent(self, async_client: AsyncClient, owner_auth_header: dict, agent: str):
        response = await async_client.put(
            self.url(uuid4()),
            headers=owner_auth_header,
            json={
                "question_id": str(uuid4()),
                "type": SuggestionType.model,
                "value": "value",
                "agent": agent,
                "score": 1.0,
            },
        )

        assert response.status_code == 422

    async def test_with_invalid_min_length_agent(self, async_client: AsyncClient, owner_auth_header: dict):
        response = await async_client.put(
            self.url(uuid4()),
            headers=owner_auth_header,
            json={
                "question_id": str(uuid4()),
                "type": SuggestionType.model,
                "value": "value",
                "agent": "",
                "score": 1.0,
            },
        )

        assert response.status_code == 422

    async def test_with_invalid_max_length_agent(self, async_client: AsyncClient, owner_auth_header: dict):
        response = await async_client.put(
            self.url(uuid4()),
            headers=owner_auth_header,
            json={
                "question_id": str(uuid4()),
                "type": SuggestionType.model,
                "value": "value",
                "agent": "a" * 201,
                "score": 1.0,
            },
        )

        assert response.status_code == 422

    async def test_with_invalid_lower_score(self, async_client: AsyncClient, owner_auth_header: dict):
        response = await async_client.put(
            self.url(uuid4()),
            headers=owner_auth_header,
            json={
                "question_id": str(uuid4()),
                "type": SuggestionType.model,
                "value": "value",
                "agent": "agent",
                "score": -0.1,
            },
        )

        assert response.status_code == 422

    async def test_with_invalid_upper_score(self, async_client: AsyncClient, owner_auth_header: dict):
        response = await async_client.put(
            self.url(uuid4()),
            headers=owner_auth_header,
            json={
                "question_id": str(uuid4()),
                "type": SuggestionType.model,
                "value": "value",
                "agent": "agent",
                "score": 1.1,
            },
        )

        assert response.status_code == 422



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
import datetime
import uuid

import pytest
from argilla_server.commons.models import TaskType
from argilla_server.daos.models.datasets import BaseDatasetDB
from argilla_server.schemas.v0.datasets import CreateDatasetRequest, Dataset

from tests.pydantic_v1 import ValidationError


@pytest.mark.parametrize(
    "name",
    ["fine", "fine33", "fine_33", "fine-3-3"],
)
def test_dataset_naming_ok(name):
    request = CreateDatasetRequest(name=name, task=TaskType.token_classification)
    assert request.name == name


@pytest.mark.parametrize(
    "name",
    [
        "WrongName",
        "-wrong_name",
        "_wrong_name",
        "wrong_name-??",
        "wrong name",
        "wrong:name",
        "wrong=?name",
    ],
)
def test_dataset_naming_ko(name):
    with pytest.raises(ValidationError, match="string does not match regex"):
        CreateDatasetRequest(name=name, task=TaskType.token_classification)


@pytest.mark.parametrize(
    ("dataset", "expected_workspace"),
    [
        (BaseDatasetDB(name="ds", workspace="ws", task=TaskType.text_classification), "ws"),
        (BaseDatasetDB(name="ds", owner="owner", task=TaskType.text_classification), "owner"),
        (BaseDatasetDB(name="ds", workspace="ws", owner="owner", task=TaskType.text_classification), "ws"),
        (BaseDatasetDB(name="ds", workspace=None, owner="ws", task=TaskType.text_classification), "ws"),
    ],
)
def test_dataset_creation_sync(dataset, expected_workspace):
    assert dataset.workspace == expected_workspace
    assert dataset.owner == dataset.workspace
    assert dataset.id == f"{dataset.workspace}.{dataset.name}"


def test_dataset_creation_fails_on_no_workspace_and_owner():
    with pytest.raises(ValueError, match="Missing workspace"):
        BaseDatasetDB(task=TaskType.text_classification, name="tedb", workspace=None, owner=None)


def test_accept_create_dataset_without_created_by():
    ds = Dataset(
        name="a-dataset",
        id=uuid.uuid4(),
        task=TaskType.text_classification,
        owner="dd",
        workspace="dd",
        created_at=datetime.datetime.utcnow(),
        last_updated=datetime.datetime.utcnow(),
    )

    assert ds
    assert ds.created_by is None


def test_change_workspace_by_setting():
    dataset = BaseDatasetDB(
        name="a-dataset",
        task=TaskType.text_classification,
        workspace="workspace_name",
        created_at=datetime.datetime.utcnow(),
        last_updated=datetime.datetime.utcnow(),
    )

    assert dataset.workspace == dataset.owner
    dataset.workspace = "another_workspace"

    assert dataset.workspace == dataset.owner



#  coding=utf-8
#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from fastapi import APIRouter, Depends, Query, Request, Security

from argilla_server.apis.v0.helpers import deprecate_endpoint
from argilla_server.apis.v0.models.commons.params import CommonTaskHandlerDependencies
from argilla_server.commons.config import TaskConfig, TasksFactory
from argilla_server.models import User
from argilla_server.pydantic_v1 import BaseModel, Field
from argilla_server.security import auth
from argilla_server.services.datasets import DatasetsService
from argilla_server.services.metrics import MetricsService


class MetricInfo(BaseModel):
    id: str = Field(description="The metric id")
    name: str = Field(description="The metric name")
    description: Optional[str] = Field(default=None, description="The metric description")


@dataclass
class MetricSummaryParams:
    request: Request

    interval: Optional[float] = Query(
        default=None,
        gt=0.0,
        description="The histogram interval for histogram summaries",
    )
    size: Optional[int] = Query(
        default=None,
        ge=1,
        description="The number of terms for terminological summaries",
    )

    @property
    def parameters(self) -> Dict[str, Any]:
        """Returns dynamic metric args found in the request query params"""
        return {
            "interval": self.interval,
            "size": self.size,
            **{k: v for k, v in self.request.query_params.items() if k not in ["interval", "size"]},
        }


def configure_router(router: APIRouter, cfg: TaskConfig):
    base_metrics_endpoint = f"/{cfg.task.value}/{{name}}/metrics"
    new_base_metrics_endpoint = f"/{{name}}/{cfg.task.value}/metrics"

    @deprecate_endpoint(
        path=base_metrics_endpoint,
        new_path=new_base_metrics_endpoint,
        router_method=router.get,
        operation_id="get_dataset_metrics",
        name="get_dataset_metrics",
    )
    async def get_dataset_metrics(
        name: str,
        request_deps: CommonTaskHandlerDependencies = Depends(),
        current_user: User = Security(auth.get_current_user),
        datasets: DatasetsService = Depends(DatasetsService.get_instance),
    ) -> List[MetricInfo]:
        dataset = await datasets.find_by_name(
            user=current_user,
            name=name,
            task=cfg.task,
            workspace=request_deps.workspace,
        )

        metrics = TasksFactory.get_task_metrics(dataset.task)
        metrics = metrics.metrics if metrics else []

        return [MetricInfo.parse_obj(metric) for metric in metrics]

    @deprecate_endpoint(
        path=base_metrics_endpoint + "/{metric}:summary",
        new_path=new_base_metrics_endpoint + "/{metric}:summary",
        router_method=router.post,
        operation_id="metric_summary",
        name="metric_summary",
    )
    async def metric_summary(
        name: str,
        metric: str,
        query: cfg.query,
        metric_params: MetricSummaryParams = Depends(),
        request_deps: CommonTaskHandlerDependencies = Depends(),
        current_user: User = Security(auth.get_current_user, scopes=[]),
        datasets: DatasetsService = Depends(DatasetsService.get_instance),
        metrics: MetricsService = Depends(MetricsService.get_instance),
    ):
        dataset = await datasets.find_by_name(
            user=current_user,
            name=name,
            task=cfg.task,
            workspace=request_deps.workspace,
        )

        metric_ = TasksFactory.find_task_metric(task=cfg.task, metric_id=metric)
        record_class = TasksFactory.get_task_record(cfg.task)

        return metrics.summarize_metric(
            dataset=dataset,
            metric=metric_,
            record_class=record_class,
            query=query,
            **metric_params.parameters,
        )


router = APIRouter(tags=["Metrics"], prefix="/datasets")
for cfg in TasksFactory.get_all_configs():
    configure_router(router, cfg)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING
from uuid import uuid4

import pytest
from argilla_server.constants import API_KEY_HEADER_NAME
from argilla_server.models import User, UserRole
from fastapi.testclient import TestClient
from sqlalchemy import func, select
from sqlalchemy.orm import joinedload

from tests.factories import AdminFactory, AnnotatorFactory, OwnerFactory, UserFactory, WorkspaceFactory

if TYPE_CHECKING:
    from httpx import AsyncClient
    from sqlalchemy.ext.asyncio import AsyncSession


@pytest.mark.asyncio
async def test_me(async_client: "AsyncClient", owner: User, owner_auth_header: dict):
    response = await async_client.get("/api/me", headers=owner_auth_header)

    assert response.status_code == 200

    response_body = response.json()
    assert response_body["id"] == str(owner.id)


@pytest.mark.asyncio
async def test_me_without_authentication(async_client: "AsyncClient"):
    response = await async_client.get("/api/me")

    assert response.status_code == 401


@pytest.mark.asyncio
async def test_me_as_owner(async_client: "AsyncClient"):
    workspace_a = await WorkspaceFactory.create(name="workspace-a")
    workspace_b = await WorkspaceFactory.create(name="workspace-b")
    owner = await OwnerFactory.create(workspaces=[workspace_a, workspace_b])
    await WorkspaceFactory.create(name="workspace-c")

    response = await async_client.get("/api/me", headers={API_KEY_HEADER_NAME: owner.api_key})

    assert response.status_code == 200

    response_body = response.json()
    assert response_body["id"] == str(owner.id)
    assert response_body["workspaces"] == ["workspace-a", "workspace-b", "workspace-c"]


@pytest.mark.asyncio
async def test_me_as_admin(async_client: "AsyncClient"):
    workspace_a = await WorkspaceFactory.create(name="workspace-a")
    workspace_b = await WorkspaceFactory.create(name="workspace-b")
    admin = await AdminFactory.create(workspaces=[workspace_a, workspace_b])
    await WorkspaceFactory.create(name="workspace-c")

    response = await async_client.get("/api/me", headers={API_KEY_HEADER_NAME: admin.api_key})

    assert response.status_code == 200

    response_body = response.json()
    assert response_body["id"] == str(admin.id)
    assert response_body["workspaces"] == ["workspace-a", "workspace-b"]


@pytest.mark.asyncio
async def test_me_as_annotator(async_client: "AsyncClient", db: "AsyncSession"):
    workspace_a = await WorkspaceFactory.create(name="workspace-a")
    workspace_b = await WorkspaceFactory.create(name="workspace-b")
    annotator = await AnnotatorFactory.create(workspaces=[workspace_a, workspace_b])
    await WorkspaceFactory.create(name="workspace-c")

    response = await async_client.get("/api/me", headers={API_KEY_HEADER_NAME: annotator.api_key})

    assert response.status_code == 200

    response_body = response.json()
    assert response_body["id"] == str(annotator.id)
    assert response_body["workspaces"] == ["workspace-a", "workspace-b"]


@pytest.mark.asyncio
async def test_list_users(async_client: "AsyncClient", owner_auth_header: dict):
    await UserFactory.create(username="username-a")
    await UserFactory.create(username="username-b")

    response = await async_client.get("/api/users", headers=owner_auth_header)

    assert response.status_code == 200

    response_body = response.json()
    assert list(map(lambda user: user["username"], response_body)) == ["owner", "username-a", "username-b"]


@pytest.mark.asyncio
async def test_list_users_without_authentication(async_client: "AsyncClient"):
    response = await async_client.get("/api/users")

    assert response.status_code == 401


@pytest.mark.asyncio
async def test_list_users_as_admin(async_client: "AsyncClient", db: "AsyncSession"):
    admin = await AdminFactory.create()

    response = await async_client.get("/api/users", headers={API_KEY_HEADER_NAME: admin.api_key})

    assert response.status_code == 403


@pytest.mark.asyncio
async def test_list_users_as_annotator(async_client: "AsyncClient", db: "AsyncSession"):
    annotator = await AnnotatorFactory.create()

    response = await async_client.get("/api/users", headers={API_KEY_HEADER_NAME: annotator.api_key})

    assert response.status_code == 403


@pytest.mark.asyncio
async def test_create_user(async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict):
    user = {"first_name": "first-name", "username": "username", "password": "12345678"}

    response = await async_client.post("/api/users", headers=owner_auth_header, json=user)

    assert response.status_code == 200
    assert (await db.execute(select(func.count(User.id)))).scalar() == 2

    db_user = (await db.execute(select(User).where(User.username == "username"))).scalar_one_or_none()
    assert db_user

    response_body = response.json()
    assert response_body["username"] == "username"
    assert response_body["api_key"] == db_user.api_key
    assert response_body["role"] == UserRole.annotator.value


@pytest.mark.asyncio
async def test_create_user_with_workspaces(async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict):
    workspace_a = await WorkspaceFactory.create(name="workspace-a")
    workspace_b = await WorkspaceFactory.create(name="workspace-b")

    user = {
        "first_name": "first-name",
        "username": "username",
        "password": "12345678",
        "workspaces": ["workspace-a", "workspace-b"],
    }
    response = await async_client.post("/api/users", headers=owner_auth_header, json=user)

    assert response.status_code == 200
    assert (await db.execute(select(func.count(User.id)))).scalar() == 2

    db_user = (
        (await db.execute(select(User).where(User.username == "username").options(joinedload(User.workspaces))))
        .unique()
        .scalar_one_or_none()
    )
    assert db_user
    assert workspace_a in db_user.workspaces
    assert workspace_b in db_user.workspaces

    response_body = response.json()
    assert response_body["username"] == "username"
    assert response_body["api_key"] == db_user.api_key
    assert response_body["role"] == UserRole.annotator.value
    assert response_body["workspaces"] == ["workspace-a", "workspace-b"]


@pytest.mark.asyncio
async def test_create_user_with_non_existent_workspaces(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    user = {
        "first_name": "first-name",
        "username": "username",
        "password": "12345678",
        "workspaces": ["i do not exist"],
    }
    response = await async_client.post("/api/users", headers=owner_auth_header, json=user)

    assert response.status_code == 422
    assert (await db.execute(select(func.count(User.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_create_user_with_non_default_role(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    user = {"first_name": "first-name", "username": "username", "password": "12345678", "role": UserRole.owner.value}

    response = await async_client.post("/api/users", headers=owner_auth_header, json=user)

    assert response.status_code == 200
    assert (await db.execute(select(func.count(User.id)))).scalar() == 2

    db_user = (await db.execute(select(User).where(User.username == "username"))).scalar_one_or_none()
    assert db_user

    response_body = response.json()
    assert response_body["username"] == "username"
    assert response_body["role"] == UserRole.owner.value


@pytest.mark.asyncio
async def test_create_user_without_authentication(async_client: "AsyncClient", db: "AsyncSession"):
    user = {"first_name": "first-name", "username": "username", "password": "12345678"}

    response = await async_client.post("/api/users", json=user)

    assert response.status_code == 401
    assert (await db.execute(select(func.count(User.id)))).scalar() == 0


@pytest.mark.asyncio
async def test_create_user_as_admin(async_client: "AsyncClient", db: "AsyncSession"):
    admin = await AdminFactory.create()
    user = {"first_name": "first-name", "username": "username", "password": "12345678"}

    response = await async_client.post("/api/users", headers={API_KEY_HEADER_NAME: admin.api_key}, json=user)

    assert response.status_code == 403
    assert (await db.execute(select(func.count(User.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_create_user_as_annotator(async_client: "AsyncClient", db: "AsyncSession"):
    annotator = await AnnotatorFactory.create()
    user = {"first_name": "first-name", "username": "username", "password": "12345678"}

    response = await async_client.post("/api/users", headers={API_KEY_HEADER_NAME: annotator.api_key}, json=user)

    assert response.status_code == 403
    assert (await db.execute(select(func.count(User.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_create_user_with_existent_username(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    await UserFactory.create(username="username")
    user = {"first_name": "first-name", "username": "username", "password": "12345678"}

    response = await async_client.post("/api/users", headers=owner_auth_header, json=user)

    assert response.status_code == 409
    assert (await db.execute(select(func.count(User.id)))).scalar() == 2


@pytest.mark.asyncio
async def test_create_user_with_invalid_min_length_first_name(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    user = {"first_name": "", "username": "username", "password": "12345678"}

    response = await async_client.post("/api/users", headers=owner_auth_header, json=user)

    assert response.status_code == 422
    assert (await db.execute(select(func.count(User.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_create_user_with_invalid_min_length_last_name(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header
):
    user = {"first_name": "first-name", "last_name": "", "username": "username", "password": "12345678"}

    response = await async_client.post("/api/users", headers=owner_auth_header, json=user)

    assert response.status_code == 422
    assert (await db.execute(select(func.count(User.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_create_user_with_invalid_username(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    user = {"first_name": "first-name", "username": "invalid username", "password": "12345678"}

    response = await async_client.post("/api/users", headers=owner_auth_header, json=user)

    assert response.status_code == 422
    assert (await db.execute(select(func.count(User.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_create_user_with_invalid_role(async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict):
    user = {"first_name": "first-name", "username": "username", "password": "12345678", "role": "invalid role"}

    response = await async_client.post("/api/users", headers=owner_auth_header, json=user)

    assert response.status_code == 422
    assert (await db.execute(select(func.count(User.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_create_user_with_invalid_min_password_length(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    user = {"first_name": "first-name", "username": "username", "password": "1234"}

    response = await async_client.post("/api/users", headers=owner_auth_header, json=user)

    assert response.status_code == 422
    assert (await db.execute(select(func.count(User.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_create_user_with_invalid_max_password_length(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    user = {"first_name": "first-name", "username": "username", "password": "p" * 101}

    response = await async_client.post("/api/users", headers=owner_auth_header, json=user)

    assert response.status_code == 422
    assert (await db.execute(select(func.count(User.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_delete_user(async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict):
    user = await UserFactory.create()

    response = await async_client.delete(f"/api/users/{user.id}", headers=owner_auth_header)

    assert response.status_code == 200
    assert (await db.execute(select(func.count(User.id)))).scalar() == 1

    response_body = response.json()
    assert response_body["id"] == str(user.id)


@pytest.mark.asyncio
async def test_delete_user_without_authentication(async_client: "AsyncClient", db: "AsyncSession"):
    user = await UserFactory.create()

    response = await async_client.delete(f"/api/users/{user.id}")

    assert response.status_code == 401
    assert (await db.execute(select(func.count(User.id)))).scalar() == 1


@pytest.mark.asyncio
async def test_delete_user_as_admin(async_client: "AsyncClient", db: "AsyncSession"):
    admin = await AdminFactory.create()
    user = await UserFactory.create()

    response = await async_client.delete(f"/api/users/{user.id}", headers={API_KEY_HEADER_NAME: admin.api_key})

    assert response.status_code == 403
    assert (await db.execute(select(func.count(User.id)))).scalar() == 2


@pytest.mark.asyncio
async def test_delete_user_as_annotator(async_client: "AsyncClient", db: "AsyncSession"):
    annotator = await AnnotatorFactory.create()
    user = await UserFactory.create()

    response = await async_client.delete(f"/api/users/{user.id}", headers={API_KEY_HEADER_NAME: annotator.api_key})

    assert response.status_code == 403
    assert (await db.execute(select(func.count(User.id)))).scalar() == 2


@pytest.mark.asyncio
async def test_delete_user_with_nonexistent_user_id(
    async_client: "AsyncClient", db: "AsyncSession", owner_auth_header: dict
):
    response = await async_client.delete(f"/api/users/{uuid4()}", headers=owner_auth_header)

    assert response.status_code == 404
    assert (await db.execute(select(func.count(User.id)))).scalar() == 1



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import uuid
from unittest.mock import MagicMock

import pytest
from argilla_server import telemetry
from argilla_server.commons.models import TaskType
from argilla_server.enums import UserRole
from argilla_server.models import User
from argilla_server.telemetry import TelemetryClient, get_telemetry_client
from fastapi import Request

mock_request = Request(scope={"type": "http", "headers": {}})


def test_disable_telemetry():
    telemetry_client = TelemetryClient(enable_telemetry=False)

    assert telemetry_client.client is None


@pytest.mark.asyncio
async def test_track_login(test_telemetry: MagicMock):
    user = User(id=uuid.uuid4(), username="argilla")
    await telemetry.track_login(request=mock_request, user=user)

    current_server_id = get_telemetry_client().server_id
    expected_event_data = {
        "accept-language": None,
        "is_default_user": True,
        "user_id": str(user.id),
        "user-agent": None,
        "user_hash": str(uuid.uuid5(current_server_id, name="argilla")),
    }
    test_telemetry.track_data.assert_called_once_with(action="UserInfoRequested", data=expected_event_data)


@pytest.mark.asyncio
async def test_track_bulk(test_telemetry):
    task, records = TaskType.token_classification, 100

    await telemetry.track_bulk(task=task, records=records)
    test_telemetry.track_data.assert_called_once_with(
        action="LogRecordsRequested", data={"task": task, "records": records}
    )


@pytest.mark.parametrize("is_oauth", [True, False])
@pytest.mark.parametrize("username", ["argilla", "john"])
def test_user_created(test_telemetry, username: str, is_oauth: bool):
    user = User(id=uuid.uuid4(), username=username, role=UserRole.owner)

    telemetry.track_user_created(user=user, is_oauth=is_oauth)
    test_telemetry.track_data.assert_called_once_with(
        action="UserCreated",
        data={
            "is_default_user": username == "argilla",
            "is_oauth": is_oauth,
            "role": user.role,
            "user_id": str(user.id),
        },
    )



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TYPE_CHECKING, Any

from rich.console import Console
from rich.panel import Panel
from rich.table import Table

if TYPE_CHECKING:
    from rich.console import RenderableType

# TODO: update colors after consulting it with UI expert
_ARGILLA_BORDER_STYLE = "red"


def get_argilla_themed_table(title: str, **kwargs: Any) -> Table:
    return Table(title=title, border_style=_ARGILLA_BORDER_STYLE, **kwargs)


def get_argilla_themed_panel(renderable: "RenderableType", title: str, success: bool = True, **kwargs: Any) -> Panel:
    if success:
        title = f"[green]{title}"
    return Panel(renderable=renderable, border_style=_ARGILLA_BORDER_STYLE, title=title, **kwargs)


def echo_in_panel(renderable: "RenderableType", title: str, success: bool = True, **kwargs: Any) -> None:
    Console().print(get_argilla_themed_panel(renderable, title, success, **kwargs))



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
from typing import Type

import pytest
from argilla_server.apis.v0.models.commons.model import BaseRecord
from argilla_server.apis.v0.models.text2text import (
    Text2TextBulkRequest,
    Text2TextRecord,
)
from argilla_server.apis.v0.models.text_classification import (
    TextClassificationBulkRequest,
    TextClassificationRecord,
)
from argilla_server.apis.v0.models.token_classification import (
    TokenClassificationBulkRequest,
    TokenClassificationRecord,
)
from argilla_server.commons.models import TaskType
from argilla_server.models import User
from httpx import AsyncClient

from tests.factories import WorkspaceFactory


async def create_dataset(async_client: "AsyncClient", task: TaskType, workspace_name: str):
    dataset = "test-dataset"
    record_text = "This is my text"
    response = await async_client.delete(f"/api/datasets/{dataset}", params={"workspace": workspace_name})
    assert response.status_code in [200, 404]

    response = await async_client.post(
        "/api/datasets", json={"name": dataset, "workspace": workspace_name, "task": task.value}
    )
    assert response.status_code == 200
    tags = {"env": "test", "type": task}

    if task == TaskType.text_classification:
        bulk_data = TextClassificationBulkRequest(
            tags=tags,
            records=[
                TextClassificationRecord(
                    **{
                        "id": 0,
                        "inputs": {"data": "my data"},
                        "prediction": {
                            "agent": "test",
                            "labels": [
                                {"class": "Test", "score": 0.3},
                                {"class": "Mocking", "score": 0.7},
                            ],
                        },
                    }
                )
            ],
        )
    elif task == TaskType.token_classification:
        bulk_data = TokenClassificationBulkRequest(
            tags=tags,
            records=[
                TokenClassificationRecord.parse_obj(
                    {
                        "id": 0,
                        "text": record_text,
                        "tokens": record_text.split(),
                    }
                )
            ],
        )
    else:
        bulk_data = Text2TextBulkRequest(
            tags=tags,
            records=[
                Text2TextRecord.parse_obj(
                    {
                        "id": 0,
                        "text": record_text,
                    }
                )
            ],
        )

    response = await async_client.post(
        f"/api/datasets/{dataset}/{task}:bulk", json=bulk_data.dict(by_alias=True), params={"workspace": workspace_name}
    )

    assert response.status_code == 200
    return dataset


@pytest.mark.parametrize(
    ("task", "expected_record_class"),
    [
        (TaskType.token_classification, TokenClassificationRecord),
        (TaskType.text_classification, TextClassificationRecord),
        (TaskType.text2text, Text2TextRecord),
    ],
)
@pytest.mark.asyncio
async def test_get_record_by_id(
    async_client: "AsyncClient",
    owner_auth_header: dict,
    task: TaskType,
    expected_record_class: Type[BaseRecord],
):
    workspace = await WorkspaceFactory.create()
    async_client.headers.update(owner_auth_header)
    dataset = await create_dataset(async_client, task=task, workspace_name=workspace.name)

    record_id = 0
    response = await async_client.get(
        f"/api/datasets/{dataset}/records/{record_id}", params={"workspace": workspace.name}
    )

    assert response.status_code == 200
    record = expected_record_class.parse_obj(response.json())
    assert record.id == record_id


@pytest.mark.parametrize(
    "task",
    [TaskType.token_classification, TaskType.text_classification, TaskType.text2text],
)
@pytest.mark.asyncio
async def test_get_record_by_id_not_found(async_client: "AsyncClient", owner_auth_header: dict, task: TaskType):
    async_client.headers.update(owner_auth_header)
    workspace = await WorkspaceFactory.create()
    dataset = await create_dataset(async_client, task=task, workspace_name=workspace.name)

    record_id = "not-found"
    response = await async_client.get(
        f"/api/datasets/{dataset}/records/{record_id}", params={"workspace": workspace.name}
    )

    assert response.status_code == 404
    assert response.json() == {
        "detail": {
            "code": "argilla.api.errors::RecordNotFound",
            "params": {
                "dataset": f"{workspace.name}.test-dataset",
                "id": record_id,
                "name": f"{workspace.name}.test-dataset.not-found",
                "type": "Record",
            },
        }
    }



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List
from uuid import UUID

from fastapi import APIRouter, Depends, Security
from sqlalchemy.ext.asyncio import AsyncSession

from argilla_server.contexts import accounts
from argilla_server.database import get_async_db
from argilla_server.errors import EntityAlreadyExistsError, EntityNotFoundError
from argilla_server.policies import WorkspacePolicy, WorkspaceUserPolicy, authorize
from argilla_server.pydantic_v1 import parse_obj_as
from argilla_server.schemas.v0.users import User
from argilla_server.schemas.v0.workspaces import Workspace, WorkspaceCreate, WorkspaceUserCreate
from argilla_server.security import auth

router = APIRouter(tags=["workspaces"])


@router.get("/workspaces", response_model=List[Workspace], response_model_exclude_none=True)
async def list_workspaces(
    *, db: AsyncSession = Depends(get_async_db), current_user: User = Security(auth.get_current_user)
):
    await authorize(current_user, WorkspacePolicy.list)

    workspaces = await accounts.list_workspaces(db)

    return parse_obj_as(List[Workspace], workspaces)


@router.post("/workspaces", response_model=Workspace, response_model_exclude_none=True)
async def create_workspace(
    *,
    db: AsyncSession = Depends(get_async_db),
    workspace_create: WorkspaceCreate,
    current_user: User = Security(auth.get_current_user),
):
    await authorize(current_user, WorkspacePolicy.create)

    if await accounts.get_workspace_by_name(db, workspace_create.name):
        raise EntityAlreadyExistsError(name=workspace_create.name, type=Workspace)

    workspace = await accounts.create_workspace(db, workspace_create)

    return Workspace.from_orm(workspace)


@router.get("/workspaces/{workspace_id}/users", response_model=List[User], response_model_exclude_none=True)
async def list_workspace_users(
    *,
    db: AsyncSession = Depends(get_async_db),
    workspace_id: UUID,
    current_user: User = Security(auth.get_current_user),
):
    await authorize(current_user, WorkspaceUserPolicy.list(workspace_id))

    workspace = await accounts.get_workspace_by_id(db, workspace_id)
    if not workspace:
        raise EntityNotFoundError(name=str(workspace_id), type=Workspace)

    await workspace.awaitable_attrs.users
    for user in workspace.users:
        await user.awaitable_attrs.workspaces
    return parse_obj_as(List[User], workspace.users)


@router.post("/workspaces/{workspace_id}/users/{user_id}", response_model=User, response_model_exclude_none=True)
async def create_workspace_user(
    *,
    db: AsyncSession = Depends(get_async_db),
    workspace_id: UUID,
    user_id: UUID,
    current_user: User = Security(auth.get_current_user),
):
    await authorize(current_user, WorkspaceUserPolicy.create)

    workspace = await accounts.get_workspace_by_id(db, workspace_id)
    if not workspace:
        raise EntityNotFoundError(name=str(workspace_id), type=Workspace)

    user = await accounts.get_user_by_id(db, user_id)
    if not user:
        raise EntityNotFoundError(name=str(user_id), type=User)

    workspace_user = await accounts.get_workspace_user_by_workspace_id_and_user_id(db, workspace_id, user_id)
    if workspace_user is not None:
        raise EntityAlreadyExistsError(name=str(user_id), type=User)

    workspace_user = await accounts.create_workspace_user(
        db, WorkspaceUserCreate(workspace_id=workspace_id, user_id=user_id)
    )
    await db.refresh(user, attribute_names=["workspaces"])

    return User.from_orm(workspace_user.user)


@router.delete("/workspaces/{workspace_id}/users/{user_id}", response_model=User, response_model_exclude_none=True)
async def delete_workspace_user(
    *,
    db: AsyncSession = Depends(get_async_db),
    workspace_id: UUID,
    user_id: UUID,
    current_user: User = Security(auth.get_current_user),
):
    workspace_user = await accounts.get_workspace_user_by_workspace_id_and_user_id(db, workspace_id, user_id)
    if not workspace_user:
        raise EntityNotFoundError(name=str(user_id), type=User)

    await authorize(current_user, WorkspaceUserPolicy.delete(workspace_user))

    user = await workspace_user.awaitable_attrs.user
    await accounts.delete_workspace_user(db, workspace_user)
    await db.refresh(user, attribute_names=["workspaces"])

    return User.from_orm(user)



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import asyncio
from typing import Union

from sqlalchemy import orm
from sqlalchemy.ext.asyncio import async_scoped_session, async_sessionmaker

task: Union[asyncio.Task, None] = None


def set_task(t: asyncio.Task):
    global task
    task = t


def get_task() -> asyncio.Task:
    return task


TestSession = async_scoped_session(async_sessionmaker(expire_on_commit=False, future=True), get_task)
SyncTestSession = orm.scoped_session(orm.sessionmaker(class_=orm.Session, expire_on_commit=False))



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import dataclasses
import logging
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from argilla_server.daos.backend.query_helpers import aggregations
from argilla_server.helpers import unflatten_dict

if TYPE_CHECKING:
    from argilla_server.daos.backend.client_adapters.base import IClientAdapter


_LOGGER = logging.getLogger(__file__)


@dataclasses.dataclass
class ElasticsearchMetric:
    id: str

    @property
    def metric_arg_names(self):
        return self.__args__

    def __post_init__(self):
        self.__args__ = self.get_function_arg_names(self._build_aggregation)

    @staticmethod
    def get_function_arg_names(func):
        return func.__code__.co_varnames

    def aggregation_request(self, *args, **kwargs) -> Optional[Union[Dict[str, Any], List[Dict[str, Any]]]]:
        """
        Configures the summary es aggregation definition
        """
        try:
            return {self.id: self._build_aggregation(*args, **kwargs)}
        except TypeError as ex:
            _LOGGER.warning(f"Cannot build metric for metric {self.id}. Error: {ex}. Skipping...")

    def aggregation_result(self, aggregation_result: Dict[str, Any]) -> Dict[str, Any]:
        """
        Parse the es aggregation result. Override this method
        for result customization

        Parameters
        ----------
        aggregation_result:
            Retrieved es aggregation result

        """
        return aggregation_result.get(self.id, aggregation_result)

    def _build_aggregation(self, *args, **kwargs) -> Dict[str, Any]:
        raise NotImplementedError()


@dataclasses.dataclass
class NestedPathElasticsearchMetric(ElasticsearchMetric):
    """
    A ``ElasticsearchMetric`` which need nested fields for summary calculation.

    Aggregations for nested fields need some extra configuration and this class
    encapsulate these common logic.

    Attributes:
    -----------
    nested_path:
        The nested
    """

    nested_path: str

    def __post_init__(self):
        self.__args__ = self.get_function_arg_names(self._inner_aggregation)

    def _inner_aggregation(self, *args, **kwargs) -> Dict[str, Any]:
        """The specific aggregation definition"""
        raise NotImplementedError()

    def _build_aggregation(self, *args, **kwargs) -> Dict[str, Any]:
        """Implements the common mechanism to define aggregations with nested fields"""
        return aggregations.nested_aggregation(
            nested_path=self.nested_path,
            inner_aggregation=self._inner_aggregation(*args, **kwargs),
        )

    def compound_nested_field(self, inner_field: str) -> str:
        return f"{self.nested_path}.{inner_field}"


@dataclasses.dataclass
class HistogramAggregation(ElasticsearchMetric):
    """
    Base elasticsearch histogram aggregation metric

    Attributes
    ----------
    field:
        The histogram field
    script:
        If provided, it will be used as scripted field
        for aggregation
    fixed_interval:
        If provided, it will used ALWAYS as the histogram
        aggregation interval
    """

    field: str
    script: Optional[Union[str, Dict[str, Any]]] = None
    fixed_interval: Optional[float] = None

    def _build_aggregation(self, interval: Optional[float] = None) -> Dict[str, Any]:
        if self.fixed_interval:
            interval = self.fixed_interval

        return aggregations.histogram_aggregation(field_name=self.field, script=self.script, interval=interval)


@dataclasses.dataclass
class TermsAggregation(ElasticsearchMetric):
    field: str = None
    script: Union[str, Dict[str, Any]] = None
    fixed_size: Optional[int] = None
    default_size: Optional[int] = None
    missing: Optional[str] = None

    def _build_aggregation(self, size: int = None) -> Dict[str, Any]:
        if self.fixed_size:
            size = self.fixed_size
        return aggregations.terms_aggregation(
            self.field,
            script=self.script,
            size=size or self.default_size,
            missing=self.missing,
        )


@dataclasses.dataclass
class BidimensionalTermsAggregation(ElasticsearchMetric):
    field_x: str
    field_y: str

    def _build_aggregation(self, size: int = None) -> Dict[str, Any]:
        return aggregations.bidimentional_terms_aggregations(
            field_name_x=self.field_x,
            field_name_y=self.field_y,
            size=size,
        )


@dataclasses.dataclass
class NestedTermsAggregation(NestedPathElasticsearchMetric):
    terms: TermsAggregation

    def __post_init__(self):
        super().__post_init__()
        self.terms.field = f"{self.nested_path}.{self.terms.field}"

    def _inner_aggregation(self, size: int) -> Dict[str, Any]:
        return self.terms.aggregation_request(size)


@dataclasses.dataclass
class NestedBidimensionalTermsAggregation(NestedPathElasticsearchMetric):
    biterms: BidimensionalTermsAggregation

    def __post_init__(self):
        super().__post_init__()
        self.biterms.field_x = f"{self.nested_path}.{self.biterms.field_x}"
        self.biterms.field_y = f"{self.nested_path}.{self.biterms.field_y}"

    def _inner_aggregation(self, size: int = None) -> Dict[str, Any]:
        return self.biterms.aggregation_request(size)


@dataclasses.dataclass
class NestedHistogramAggregation(NestedPathElasticsearchMetric):
    histogram: HistogramAggregation

    def __post_init__(self):
        super().__post_init__()
        self.histogram.field = f"{self.nested_path}.{self.histogram.field}"

    def _inner_aggregation(self, interval: float) -> Dict[str, Any]:
        return self.histogram.aggregation_request(interval)


@dataclasses.dataclass
class WordCloudAggregation(ElasticsearchMetric):
    default_field: str

    DEFAULT_WORDCOUNT_SIZE = 80

    def _build_aggregation(
        self,
        text_field: str = None,
        size: int = None,
    ) -> Dict[str, Any]:
        field = text_field or self.default_field
        terms_id = f"{self.id}_{field}" if text_field else self.id
        return TermsAggregation(
            id=terms_id,
            field=field,
        ).aggregation_request(
            size=size or self.DEFAULT_WORDCOUNT_SIZE
        )[terms_id]


@dataclasses.dataclass
class MetadataAggregations(ElasticsearchMetric):
    def __post_init__(self):
        super().__post_init__()
        self.__args__ = self.get_function_arg_names(self.aggregation_request)

    def aggregation_request(
        self,
        client: "IClientAdapter",
        index: str,
        size: int = None,
    ) -> List[Dict[str, Any]]:
        schema = client.get_property_type(
            index=index,
            property_name="metadata",
        )

        metadata_aggs = aggregations.custom_fields(
            fields_definitions=schema,
            size=size,
        )
        return [{key: value} for key, value in metadata_aggs.items()]

    def aggregation_result(self, aggregation_result: Dict[str, Any]) -> Dict[str, Any]:
        data = unflatten_dict(aggregation_result, stop_keys=["metadata"])
        return data.get("metadata", {})



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import List, Optional, Tuple, Type, Union
from uuid import uuid4

import pytest
from argilla_server.apis.v1.handlers.datasets.records import LIST_DATASET_RECORDS_LIMIT_DEFAULT
from argilla_server.constants import API_KEY_HEADER_NAME
from argilla_server.enums import RecordInclude, RecordSortField, ResponseStatus, UserRole
from argilla_server.models import Dataset, Question, Record, Response, Suggestion, User, Workspace
from argilla_server.search_engine import (
    FloatMetadataFilter,
    IntegerMetadataFilter,
    MetadataFilter,
    SearchEngine,
    SearchResponseItem,
    SearchResponses,
    SortBy,
    TermsMetadataFilter,
)
from httpx import AsyncClient

from tests.factories import (
    AdminFactory,
    AnnotatorFactory,
    DatasetFactory,
    LabelSelectionQuestionFactory,
    MetadataPropertyFactory,
    RecordFactory,
    ResponseFactory,
    SuggestionFactory,
    TermsMetadataPropertyFactory,
    TextFieldFactory,
    TextQuestionFactory,
    UserFactory,
    VectorFactory,
    VectorSettingsFactory,
    WorkspaceFactory,
)


@pytest.mark.asyncio
class TestSuiteListDatasetRecords:
    @pytest.mark.skip(reason="Factory integration with search engine")
    async def test_list_dataset_records(self, async_client: "AsyncClient", owner_auth_header: dict):
        dataset = await DatasetFactory.create()
        record_a = await RecordFactory.create(fields={"record_a": "value_a"}, dataset=dataset)
        record_b = await RecordFactory.create(
            fields={"record_b": "value_b"}, metadata_={"unit": "test"}, dataset=dataset
        )
        record_c = await RecordFactory.create(fields={"record_c": "value_c"}, dataset=dataset)

        other_dataset = await DatasetFactory.create()
        await RecordFactory.create_batch(size=2, dataset=other_dataset)

        response = await async_client.get(f"/api/v1/datasets/{dataset.id}/records", headers=owner_auth_header)

        assert response.status_code == 200
        assert response.json() == {
            "total": 3,
            "items": [
                {
                    "id": str(record_a.id),
                    "fields": {"record_a": "value_a"},
                    "metadata": None,
                    "external_id": record_a.external_id,
                    "inserted_at": record_a.inserted_at.isoformat(),
                    "updated_at": record_a.updated_at.isoformat(),
                },
                {
                    "id": str(record_b.id),
                    "fields": {"record_b": "value_b"},
                    "metadata": {"unit": "test"},
                    "external_id": record_b.external_id,
                    "inserted_at": record_b.inserted_at.isoformat(),
                    "updated_at": record_b.updated_at.isoformat(),
                },
                {
                    "id": str(record_c.id),
                    "fields": {"record_c": "value_c"},
                    "metadata": None,
                    "external_id": record_c.external_id,
                    "inserted_at": record_c.inserted_at.isoformat(),
                    "updated_at": record_c.updated_at.isoformat(),
                },
            ],
        }

    @pytest.mark.parametrize(
        "includes",
        [[RecordInclude.responses], [RecordInclude.suggestions], [RecordInclude.responses, RecordInclude.suggestions]],
    )
    async def test_list_dataset_records_with_include(
        self, async_client: "AsyncClient", owner: User, owner_auth_header: dict, includes: List[RecordInclude]
    ):
        workspace = await WorkspaceFactory.create()
        dataset, questions, records, responses, suggestions = await self.create_dataset_with_user_responses(
            owner, workspace
        )
        record_a, record_b, record_c = records
        response_a_user, response_b_user = responses[1], responses[3]
        suggestion_a, suggestion_b = suggestions

        other_dataset = await DatasetFactory.create()
        await RecordFactory.create_batch(size=2, dataset=other_dataset)

        expected = {
            "total": len(records),
            "items": [
                {
                    "id": str(record_a.id),
                    "fields": {"input": "value_a"},
                    "metadata": None,
                    "external_id": record_a.external_id,
                    "inserted_at": record_a.inserted_at.isoformat(),
                    "updated_at": record_a.updated_at.isoformat(),
                },
                {
                    "id": str(record_b.id),
                    "fields": {"input": "value_b"},
                    "metadata": {"unit": "test"},
                    "external_id": record_b.external_id,
                    "inserted_at": record_b.inserted_at.isoformat(),
                    "updated_at": record_b.updated_at.isoformat(),
                },
                {
                    "id": str(record_c.id),
                    "fields": {"input": "value_c"},
                    "metadata": None,
                    "external_id": record_c.external_id,
                    "inserted_at": record_c.inserted_at.isoformat(),
                    "updated_at": record_c.updated_at.isoformat(),
                },
            ],
        }

        if RecordInclude.responses in includes:
            expected["items"][0]["responses"] = [
                {
                    "id": str(response_a_user.id),
                    "values": None,
                    "status": "discarded",
                    "user_id": str(owner.id),
                    "inserted_at": response_a_user.inserted_at.isoformat(),
                    "updated_at": response_a_user.updated_at.isoformat(),
                }
            ]
            expected["items"][1]["responses"] = [
                {
                    "id": str(response_b_user.id),
                    "values": {
                        "input_ok": {"value": "no"},
                        "output_ok": {"value": "no"},
                    },
                    "status": "submitted",
                    "user_id": str(owner.id),
                    "inserted_at": response_b_user.inserted_at.isoformat(),
                    "updated_at": response_b_user.updated_at.isoformat(),
                },
            ]
            expected["items"][2]["responses"] = []

        if RecordInclude.suggestions in includes:
            expected["items"][0]["suggestions"] = [
                {
                    "id": str(suggestion_a.id),
                    "value": "option-1",
                    "score": None,
                    "agent": None,
                    "type": None,
                    "question_id": str(questions[0].id),
                }
            ]
            expected["items"][1]["suggestions"] = [
                {
                    "id": str(suggestion_b.id),
                    "value": "option-2",
                    "score": 0.75,
                    "agent": "unit-test-agent",
                    "type": "model",
                    "question_id": str(questions[0].id),
                }
            ]
            expected["items"][2]["suggestions"] = []

        response = await async_client.get(
            f"/api/v1/datasets/{dataset.id}/records",
            params={"include": RecordInclude.responses.value},
            headers=owner_auth_header,
        )

        assert response.status_code == 200

    @pytest.mark.skip(reason="Factory integration with search engine")
    async def test_list_dataset_records_with_include_vectors(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()
        record_a = await RecordFactory.create(dataset=dataset)
        record_b = await RecordFactory.create(dataset=dataset)
        record_c = await RecordFactory.create(dataset=dataset)
        vector_settings_a = await VectorSettingsFactory.create(name="vector-a", dimensions=3, dataset=dataset)
        vector_settings_b = await VectorSettingsFactory.create(name="vector-b", dimensions=2, dataset=dataset)

        await VectorFactory.create(value=[1.0, 2.0, 3.0], vector_settings=vector_settings_a, record=record_a)
        await VectorFactory.create(value=[4.0, 5.0], vector_settings=vector_settings_b, record=record_a)
        await VectorFactory.create(value=[1.0, 2.0], vector_settings=vector_settings_b, record=record_b)

        response = await async_client.get(
            f"/api/v1/datasets/{dataset.id}/records",
            params={"include": RecordInclude.vectors.value},
            headers=owner_auth_header,
        )

        assert response.status_code == 200
        assert response.json() == {
            "items": [
                {
                    "id": str(record_a.id),
                    "fields": {"text": "This is a text", "sentiment": "neutral"},
                    "metadata": None,
                    "external_id": record_a.external_id,
                    "vectors": {
                        "vector-a": [1.0, 2.0, 3.0],
                        "vector-b": [4.0, 5.0],
                    },
                    "inserted_at": record_a.inserted_at.isoformat(),
                    "updated_at": record_a.updated_at.isoformat(),
                },
                {
                    "id": str(record_b.id),
                    "fields": {"text": "This is a text", "sentiment": "neutral"},
                    "metadata": None,
                    "external_id": record_b.external_id,
                    "vectors": {
                        "vector-b": [1.0, 2.0],
                    },
                    "inserted_at": record_b.inserted_at.isoformat(),
                    "updated_at": record_b.updated_at.isoformat(),
                },
                {
                    "id": str(record_c.id),
                    "fields": {"text": "This is a text", "sentiment": "neutral"},
                    "metadata": None,
                    "external_id": record_c.external_id,
                    "vectors": {},
                    "inserted_at": record_c.inserted_at.isoformat(),
                    "updated_at": record_c.updated_at.isoformat(),
                },
            ],
            "total": 3,
        }

    @pytest.mark.skip(reason="Factory integration with search engine")
    async def test_list_dataset_records_with_include_specific_vectors(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()
        record_a = await RecordFactory.create(dataset=dataset)
        record_b = await RecordFactory.create(dataset=dataset)
        record_c = await RecordFactory.create(dataset=dataset)
        vector_settings_a = await VectorSettingsFactory.create(name="vector-a", dimensions=3, dataset=dataset)
        vector_settings_b = await VectorSettingsFactory.create(name="vector-b", dimensions=2, dataset=dataset)
        vector_settings_c = await VectorSettingsFactory.create(name="vector-c", dimensions=4, dataset=dataset)

        await VectorFactory.create(value=[1.0, 2.0, 3.0], vector_settings=vector_settings_a, record=record_a)
        await VectorFactory.create(value=[4.0, 5.0], vector_settings=vector_settings_b, record=record_a)
        await VectorFactory.create(value=[6.0, 7.0, 8.0, 9.0], vector_settings=vector_settings_c, record=record_a)
        await VectorFactory.create(value=[1.0, 2.0], vector_settings=vector_settings_b, record=record_b)
        await VectorFactory.create(value=[10.0, 11.0, 12.0, 13.0], vector_settings=vector_settings_c, record=record_b)
        await VectorFactory.create(value=[14.0, 15.0, 16.0, 17.0], vector_settings=vector_settings_c, record=record_c)

        response = await async_client.get(
            f"/api/v1/datasets/{dataset.id}/records",
            params={"include": f"{RecordInclude.vectors.value}:{vector_settings_a.name},{vector_settings_b.name}"},
            headers=owner_auth_header,
        )

        assert response.status_code == 200
        assert response.json() == {
            "items": [
                {
                    "id": str(record_a.id),
                    "fields": {"text": "This is a text", "sentiment": "neutral"},
                    "metadata": None,
                    "external_id": record_a.external_id,
                    "vectors": {
                        "vector-a": [1.0, 2.0, 3.0],
                        "vector-b": [4.0, 5.0],
                    },
                    "inserted_at": record_a.inserted_at.isoformat(),
                    "updated_at": record_a.updated_at.isoformat(),
                },
                {
                    "id": str(record_b.id),
                    "fields": {"text": "This is a text", "sentiment": "neutral"},
                    "metadata": None,
                    "external_id": record_b.external_id,
                    "vectors": {
                        "vector-b": [1.0, 2.0],
                    },
                    "inserted_at": record_b.inserted_at.isoformat(),
                    "updated_at": record_b.updated_at.isoformat(),
                },
                {
                    "id": str(record_c.id),
                    "fields": {"text": "This is a text", "sentiment": "neutral"},
                    "metadata": None,
                    "external_id": record_c.external_id,
                    "vectors": {},
                    "inserted_at": record_c.inserted_at.isoformat(),
                    "updated_at": record_c.updated_at.isoformat(),
                },
            ],
            "total": 3,
        }

    @pytest.mark.skip(reason="Factory integration with search engine")
    async def test_list_dataset_records_with_offset(self, async_client: "AsyncClient", owner_auth_header: dict):
        dataset = await DatasetFactory.create()
        await RecordFactory.create(fields={"record_a": "value_a"}, dataset=dataset)
        await RecordFactory.create(fields={"record_b": "value_b"}, dataset=dataset)
        record_c = await RecordFactory.create(fields={"record_c": "value_c"}, dataset=dataset)

        other_dataset = await DatasetFactory.create()
        await RecordFactory.create_batch(size=2, dataset=other_dataset)

        response = await async_client.get(
            f"/api/v1/datasets/{dataset.id}/records", headers=owner_auth_header, params={"offset": 2}
        )

        assert response.status_code == 200

        response_body = response.json()
        assert [item["id"] for item in response_body["items"]] == [str(record_c.id)]

    @pytest.mark.skip(reason="Factory integration with search engine")
    async def test_list_dataset_records_with_limit(self, async_client: "AsyncClient", owner_auth_header: dict):
        dataset = await DatasetFactory.create()
        record_a = await RecordFactory.create(fields={"record_a": "value_a"}, dataset=dataset)
        await RecordFactory.create(fields={"record_b": "value_b"}, dataset=dataset)
        await RecordFactory.create(fields={"record_c": "value_c"}, dataset=dataset)

        other_dataset = await DatasetFactory.create()
        await RecordFactory.create_batch(size=2, dataset=other_dataset)

        response = await async_client.get(
            f"/api/v1/datasets/{dataset.id}/records", headers=owner_auth_header, params={"limit": 1}
        )

        assert response.status_code == 200

        response_body = response.json()
        assert [item["id"] for item in response_body["items"]] == [str(record_a.id)]

    @pytest.mark.skip(reason="Factory integration with search engine")
    async def test_list_dataset_records_with_offset_and_limit(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()
        await RecordFactory.create(fields={"record_a": "value_a"}, dataset=dataset)
        record_c = await RecordFactory.create(fields={"record_b": "value_b"}, dataset=dataset)
        await RecordFactory.create(fields={"record_c": "value_c"}, dataset=dataset)

        other_dataset = await DatasetFactory.create()
        await RecordFactory.create_batch(size=2, dataset=other_dataset)

        response = await async_client.get(
            f"/api/v1/datasets/{dataset.id}/records", headers=owner_auth_header, params={"offset": 1, "limit": 1}
        )

        assert response.status_code == 200

        response_body = response.json()
        assert [item["id"] for item in response_body["items"]] == [str(record_c.id)]

    async def create_records_with_response(
        self,
        num_records: int,
        dataset: Dataset,
        user: User,
        response_status: ResponseStatus,
        response_values: Optional[dict] = None,
    ):
        for record in await RecordFactory.create_batch(size=num_records, dataset=dataset):
            await ResponseFactory.create(record=record, user=user, values=response_values, status=response_status)

    @pytest.mark.parametrize(
        ("property_config", "param_value", "expected_filter_class", "expected_filter_args"),
        [
            (
                {"name": "terms_prop", "settings": {"type": "terms"}},
                "value",
                TermsMetadataFilter,
                dict(values=["value"]),
            ),
            (
                {"name": "terms_prop", "settings": {"type": "terms"}},
                "value1,value2",
                TermsMetadataFilter,
                dict(values=["value1", "value2"]),
            ),
            (
                {"name": "integer_prop", "settings": {"type": "integer"}},
                '{"ge": 10, "le": 20}',
                IntegerMetadataFilter,
                dict(ge=10, le=20),
            ),
            (
                {"name": "integer_prop", "settings": {"type": "integer"}},
                '{"ge": 20}',
                IntegerMetadataFilter,
                dict(ge=20, high=None),
            ),
            (
                {"name": "integer_prop", "settings": {"type": "integer"}},
                '{"le": 20}',
                IntegerMetadataFilter,
                dict(ge=None, le=20),
            ),
            (
                {"name": "float_prop", "settings": {"type": "float"}},
                '{"ge": -1.30, "le": 23.23}',
                FloatMetadataFilter,
                dict(ge=-1.30, le=23.23),
            ),
            (
                {"name": "float_prop", "settings": {"type": "float"}},
                '{"ge": 23.23}',
                FloatMetadataFilter,
                dict(ge=23.23, high=None),
            ),
            (
                {"name": "float_prop", "settings": {"type": "float"}},
                '{"le": 11.32}',
                FloatMetadataFilter,
                dict(ge=None, le=11.32),
            ),
        ],
    )
    async def test_list_dataset_records_with_metadata_filter(
        self,
        async_client: "AsyncClient",
        mock_search_engine: SearchEngine,
        owner: User,
        owner_auth_header: dict,
        property_config: dict,
        param_value: str,
        expected_filter_class: Type[MetadataFilter],
        expected_filter_args: dict,
    ):
        workspace = await WorkspaceFactory.create()
        dataset, _, records, _, _ = await self.create_dataset_with_user_responses(owner, workspace)

        metadata_property = await MetadataPropertyFactory.create(
            name=property_config["name"],
            settings=property_config["settings"],
            dataset=dataset,
        )

        mock_search_engine.search.return_value = SearchResponses(
            total=2,
            items=[
                SearchResponseItem(record_id=records[0].id, score=14.2),
                SearchResponseItem(record_id=records[1].id, score=12.2),
            ],
        )

        query_params = {"metadata": [f"{metadata_property.name}:{param_value}"]}
        response = await async_client.get(
            f"/api/v1/datasets/{dataset.id}/records",
            params=query_params,
            headers=owner_auth_header,
        )
        assert response.status_code == 200

        response_json = response.json()
        assert response_json["total"] == 2

        mock_search_engine.search.assert_called_once_with(
            dataset=dataset,
            query=None,
            metadata_filters=[expected_filter_class(metadata_property=metadata_property, **expected_filter_args)],
            user_response_status_filter=None,
            offset=0,
            limit=LIST_DATASET_RECORDS_LIMIT_DEFAULT,
            sort_by=[SortBy(field=RecordSortField.inserted_at)],
        )

    @pytest.mark.skip(reason="Factory integration with search engine")
    @pytest.mark.parametrize(
        "response_status_filter", ["missing", "pending", "discarded", "submitted", "draft", ["submitted", "draft"]]
    )
    async def test_list_dataset_records_with_response_status_filter(
        self,
        async_client: "AsyncClient",
        owner: "User",
        owner_auth_header: dict,
        response_status_filter: Union[str, List[str]],
    ):
        num_records_per_response_status = 10
        response_values = {"input_ok": {"value": "yes"}, "output_ok": {"value": "yes"}}

        dataset = await DatasetFactory.create()
        # missing responses
        await RecordFactory.create_batch(size=num_records_per_response_status, dataset=dataset)
        # discarded responses
        await self.create_records_with_response(
            num_records_per_response_status, dataset, owner, ResponseStatus.discarded
        )
        # submitted responses
        await self.create_records_with_response(
            num_records_per_response_status, dataset, owner, ResponseStatus.submitted, response_values
        )
        # drafted responses
        await self.create_records_with_response(
            num_records_per_response_status, dataset, owner, ResponseStatus.draft, response_values
        )

        other_dataset = await DatasetFactory.create()
        await RecordFactory.create_batch(size=2, dataset=other_dataset)

        response_status_filter = (
            [response_status_filter] if isinstance(response_status_filter, str) else response_status_filter
        )
        response_status_filter_url = [
            f"response_status={response_status}" for response_status in response_status_filter
        ]

        response = await async_client.get(
            f"/api/v1/datasets/{dataset.id}/records?{'&'.join(response_status_filter_url)}&include=responses",
            headers=owner_auth_header,
        )

        assert response.status_code == 200
        response_json = response.json()

        assert len(response_json["items"]) == (num_records_per_response_status * len(response_status_filter))

        if "missing" in response_status_filter:
            assert (
                len([record for record in response_json["items"] if len(record["responses"]) == 0])
                >= num_records_per_response_status
            )
        assert all(
            [
                record["responses"][0]["status"] in response_status_filter
                for record in response_json["items"]
                if len(record["responses"]) > 0
            ]
        )

    @pytest.mark.parametrize(
        "sorts",
        [
            [("inserted_at", None)],
            [("inserted_at", "asc")],
            [("inserted_at", "desc")],
            [("updated_at", None)],
            [("updated_at", "asc")],
            [("updated_at", "desc")],
            [("metadata.terms-metadata-property", None)],
            [("metadata.terms-metadata-property", "asc")],
            [("metadata.terms-metadata-property", "desc")],
            [("inserted_at", "asc"), ("updated_at", "desc")],
            [("inserted_at", "desc"), ("updated_at", "asc")],
            [("inserted_at", "asc"), ("metadata.terms-metadata-property", "desc")],
            [("inserted_at", "desc"), ("metadata.terms-metadata-property", "asc")],
            [("updated_at", "asc"), ("metadata.terms-metadata-property", "desc")],
            [("updated_at", "desc"), ("metadata.terms-metadata-property", "asc")],
            [("inserted_at", "asc"), ("updated_at", "desc"), ("metadata.terms-metadata-property", "asc")],
            [("inserted_at", "desc"), ("updated_at", "asc"), ("metadata.terms-metadata-property", "desc")],
            [("inserted_at", "asc"), ("updated_at", "asc"), ("metadata.terms-metadata-property", "desc")],
        ],
    )
    async def test_list_dataset_records_with_sort_by(
        self,
        async_client: "AsyncClient",
        mock_search_engine: SearchEngine,
        owner: "User",
        owner_auth_header: dict,
        sorts: List[Tuple[str, Union[str, None]]],
    ):
        workspace = await WorkspaceFactory.create()
        dataset, _, records, _, _ = await self.create_dataset_with_user_responses(owner, workspace)

        expected_sorts_by = []
        for field, order in sorts:
            if field not in ("inserted_at", "updated_at"):
                field = await TermsMetadataPropertyFactory.create(name=field.split(".")[-1], dataset=dataset)
            expected_sorts_by.append(SortBy(field=field, order=order or "asc"))

        mock_search_engine.search.return_value = SearchResponses(
            total=2,
            items=[
                SearchResponseItem(record_id=records[0].id, score=14.2),
                SearchResponseItem(record_id=records[1].id, score=12.2),
            ],
        )

        query_params = {
            "sort_by": [f"{field}:{order}" if order is not None else f"{field}:asc" for field, order in sorts]
        }

        response = await async_client.get(
            f"/api/v1/datasets/{dataset.id}/records",
            params=query_params,
            headers=owner_auth_header,
        )
        assert response.status_code == 200
        assert response.json()["total"] == 2

        mock_search_engine.search.assert_called_once_with(
            dataset=dataset,
            query=None,
            metadata_filters=[],
            user_response_status_filter=None,
            offset=0,
            limit=LIST_DATASET_RECORDS_LIMIT_DEFAULT,
            sort_by=expected_sorts_by,
        )

    async def test_list_dataset_records_with_sort_by_with_wrong_sort_order_value(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()

        response = await async_client.get(
            f"/api/v1/datasets/{dataset.id}/records", params={"sort_by": "inserted_at:wrong"}, headers=owner_auth_header
        )
        assert response.status_code == 422
        assert response.json() == {
            "detail": "Provided sort order in 'sort_by' query param 'wrong' for field 'inserted_at' is not valid."
        }

    async def test_list_dataset_records_with_sort_by_with_non_existent_metadata_property(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()

        response = await async_client.get(
            f"/api/v1/datasets/{dataset.id}/records",
            params={"sort_by": "metadata.i-do-not-exist:asc"},
            headers=owner_auth_header,
        )
        assert response.status_code == 422
        assert response.json() == {
            "detail": f"Provided metadata property in 'sort_by' query param 'i-do-not-exist' not found in dataset with '{dataset.id}'."
        }

    async def test_list_dataset_records_with_sort_by_with_invalid_field(
        self, async_client: "AsyncClient", owner: "User", owner_auth_header: dict
    ):
        workspace = await WorkspaceFactory.create()
        dataset, _, _, _, _ = await self.create_dataset_with_user_responses(owner, workspace)

        response = await async_client.get(
            f"/api/v1/datasets/{dataset.id}/records",
            params={"sort_by": "not-valid"},
            headers=owner_auth_header,
        )
        assert response.status_code == 422
        assert response.json() == {
            "detail": "Provided sort field in 'sort_by' query param 'not-valid' is not valid. "
            "It must be either 'inserted_at', 'updated_at' or `metadata.metadata-property-name`"
        }

    async def test_list_dataset_records_without_authentication(self, async_client: "AsyncClient"):
        dataset = await DatasetFactory.create()

        response = await async_client.get(f"/api/v1/datasets/{dataset.id}/records")

        assert response.status_code == 401

    async def test_list_dataset_records_as_admin(self, async_client: "AsyncClient"):
        workspace = await WorkspaceFactory.create()
        admin = await AdminFactory.create(workspaces=[workspace])
        dataset = await DatasetFactory.create(workspace=workspace)

        await RecordFactory.create(fields={"record_a": "value_a"}, dataset=dataset)
        await RecordFactory.create(fields={"record_b": "value_b"}, dataset=dataset)
        await RecordFactory.create(fields={"record_c": "value_c"}, dataset=dataset)

        other_dataset = await DatasetFactory.create()
        await RecordFactory.create_batch(size=2, dataset=other_dataset)

        response = await async_client.get(
            f"/api/v1/datasets/{dataset.id}/records", headers={API_KEY_HEADER_NAME: admin.api_key}
        )
        assert response.status_code == 200

    async def test_list_dataset_records_as_annotator(self, async_client: "AsyncClient"):
        workspace = await WorkspaceFactory.create()
        annotator = await AnnotatorFactory.create(workspaces=[workspace])
        dataset = await DatasetFactory.create(workspace=workspace)

        await RecordFactory.create(fields={"record_a": "value_a"}, dataset=dataset)
        await RecordFactory.create(fields={"record_b": "value_b"}, dataset=dataset)
        await RecordFactory.create(fields={"record_c": "value_c"}, dataset=dataset)

        other_dataset = await DatasetFactory.create()
        await RecordFactory.create_batch(size=2, dataset=other_dataset)

        response = await async_client.get(
            f"/api/v1/datasets/{dataset.id}/records", headers={API_KEY_HEADER_NAME: annotator.api_key}
        )
        assert response.status_code == 403

    async def create_dataset_with_user_responses(
        self, user: User, workspace: Workspace
    ) -> Tuple[Dataset, List[Question], List[Record], List[Response], List[Suggestion]]:
        dataset = await DatasetFactory.create(workspace=workspace)
        await TextFieldFactory.create(name="input", dataset=dataset)
        await TextFieldFactory.create(name="output", dataset=dataset)

        annotator = await AnnotatorFactory.create(workspaces=[dataset.workspace])

        questions = [
            await LabelSelectionQuestionFactory.create(dataset=dataset),
            await TextQuestionFactory.create(name="input_ok", dataset=dataset),
            await TextQuestionFactory.create(name="output_ok", dataset=dataset),
        ]

        records = [
            await RecordFactory.create(fields={"input": "input_a", "output": "output_a"}, dataset=dataset),
            await RecordFactory.create(
                fields={"input": "input_b", "output": "output_b"}, metadata_={"unit": "test"}, dataset=dataset
            ),
            await RecordFactory.create(fields={"input": "input_c", "output": "output_c"}, dataset=dataset),
        ]

        responses = [
            await ResponseFactory.create(
                values={
                    "input_ok": {"value": "yes"},
                    "output_ok": {"value": "yes"},
                },
                record=records[0],
                user=annotator,
            ),
            await ResponseFactory.create(status="discarded", record=records[0], user=user),
            await ResponseFactory.create(
                values={
                    "input_ok": {"value": "yes"},
                    "output_ok": {"value": "no"},
                },
                record=records[1],
                user=annotator,
            ),
            await ResponseFactory.create(
                values={
                    "input_ok": {"value": "no"},
                    "output_ok": {"value": "no"},
                },
                record=records[1],
                user=user,
            ),
            await ResponseFactory.create(
                values={
                    "input_ok": {"value": "yes"},
                    "output_ok": {"value": "yes"},
                },
                record=records[1],
            ),
        ]

        # Add some responses from other users
        await ResponseFactory.create_batch(10, record=records[0], status=ResponseStatus.submitted)

        suggestions = [
            await SuggestionFactory.create(record=records[0], question=questions[0], value="option-1"),
            await SuggestionFactory.create(
                record=records[1],
                question=questions[0],
                value="option-2",
                score=0.75,
                agent="unit-test-agent",
                type="model",
            ),
        ]

        return dataset, questions, records, responses, suggestions

    @pytest.mark.skip(reason="Factory integration with search engine")
    async def test_list_current_user_dataset_records(
        self, async_client: "AsyncClient", owner: User, owner_auth_header: dict
    ):
        workspace = await WorkspaceFactory.create()
        dataset, _, records, _, _ = await self.create_dataset_with_user_responses(owner, workspace)
        record_a, record_b, record_c = records

        other_dataset = await DatasetFactory.create()
        await RecordFactory.create_batch(size=2, dataset=other_dataset)

        response = await async_client.get(f"/api/v1/me/datasets/{dataset.id}/records", headers=owner_auth_header)

        assert response.status_code == 200
        assert response.json() == {
            "total": 3,
            "items": [
                {
                    "id": str(record_a.id),
                    "fields": {"input": "input_a", "output": "output_a"},
                    "metadata": None,
                    "external_id": record_a.external_id,
                    "inserted_at": record_a.inserted_at.isoformat(),
                    "updated_at": record_a.updated_at.isoformat(),
                },
                {
                    "id": str(record_b.id),
                    "fields": {"input": "input_b", "output": "output_b"},
                    "metadata": {"unit": "test"},
                    "external_id": record_b.external_id,
                    "inserted_at": record_b.inserted_at.isoformat(),
                    "updated_at": record_b.updated_at.isoformat(),
                },
                {
                    "id": str(record_c.id),
                    "fields": {"input": "input_c", "output": "output_c"},
                    "metadata": None,
                    "external_id": record_c.external_id,
                    "inserted_at": record_c.inserted_at.isoformat(),
                    "updated_at": record_c.updated_at.isoformat(),
                },
            ],
        }

    @pytest.mark.skip(reason="Factory integration with search engine")
    @pytest.mark.parametrize("role", [UserRole.annotator, UserRole.admin, UserRole.owner])
    @pytest.mark.parametrize(
        "includes",
        [[RecordInclude.responses], [RecordInclude.suggestions], [RecordInclude.responses, RecordInclude.suggestions]],
    )
    async def test_list_current_user_dataset_records_with_include(
        self, async_client: "AsyncClient", role: UserRole, includes: List[RecordInclude]
    ):
        workspace = await WorkspaceFactory.create()
        user = await UserFactory.create(workspaces=[workspace], role=role)
        dataset, questions, records, responses, suggestions = await self.create_dataset_with_user_responses(
            user, workspace
        )
        record_a, record_b, record_c = records
        response_a_user, response_b_user = responses[1], responses[3]
        suggestion_a, suggestion_b = suggestions

        other_dataset = await DatasetFactory.create()
        await RecordFactory.create_batch(size=2, dataset=other_dataset)

        params = [("include", include.value) for include in includes]
        response = await async_client.get(
            f"/api/v1/me/datasets/{dataset.id}/records", params=params, headers={API_KEY_HEADER_NAME: user.api_key}
        )

        expected = {
            "total": 3,
            "items": [
                {
                    "id": str(record_a.id),
                    "fields": {"input": "input_a", "output": "output_a"},
                    "metadata": None,
                    "external_id": record_a.external_id,
                    "inserted_at": record_a.inserted_at.isoformat(),
                    "updated_at": record_a.updated_at.isoformat(),
                },
                {
                    "id": str(record_b.id),
                    "fields": {"input": "input_b", "output": "output_b"},
                    "metadata": {"unit": "test"},
                    "external_id": record_b.external_id,
                    "inserted_at": record_b.inserted_at.isoformat(),
                    "updated_at": record_b.updated_at.isoformat(),
                },
                {
                    "id": str(record_c.id),
                    "fields": {"input": "input_c", "output": "output_c"},
                    "metadata": None,
                    "external_id": record_c.external_id,
                    "inserted_at": record_c.inserted_at.isoformat(),
                    "updated_at": record_c.updated_at.isoformat(),
                },
            ],
        }

        if RecordInclude.responses in includes:
            expected["items"][0]["responses"] = [
                {
                    "id": str(response_a_user.id),
                    "values": None,
                    "status": "discarded",
                    "user_id": str(user.id),
                    "inserted_at": response_a_user.inserted_at.isoformat(),
                    "updated_at": response_a_user.updated_at.isoformat(),
                }
            ]
            expected["items"][1]["responses"] = [
                {
                    "id": str(response_b_user.id),
                    "values": {
                        "input_ok": {"value": "no"},
                        "output_ok": {"value": "no"},
                    },
                    "status": "submitted",
                    "user_id": str(user.id),
                    "inserted_at": response_b_user.inserted_at.isoformat(),
                    "updated_at": response_b_user.updated_at.isoformat(),
                },
            ]
            expected["items"][2]["responses"] = []

        if RecordInclude.suggestions in includes:
            expected["items"][0]["suggestions"] = [
                {
                    "id": str(suggestion_a.id),
                    "value": "option-1",
                    "score": None,
                    "agent": None,
                    "type": None,
                    "question_id": str(questions[0].id),
                }
            ]
            expected["items"][1]["suggestions"] = [
                {
                    "id": str(suggestion_b.id),
                    "value": "option-2",
                    "score": 0.75,
                    "agent": "unit-test-agent",
                    "type": "model",
                    "question_id": str(questions[0].id),
                }
            ]
            expected["items"][2]["suggestions"] = []

        assert response.status_code == 200
        assert response.json() == expected

    @pytest.mark.skip(reason="Factory integration with search engine")
    async def test_list_current_user_dataset_records_with_include_vectors(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()
        record_a = await RecordFactory.create(dataset=dataset)
        record_b = await RecordFactory.create(dataset=dataset)
        record_c = await RecordFactory.create(dataset=dataset)
        vector_settings_a = await VectorSettingsFactory.create(name="vector-a", dimensions=3, dataset=dataset)
        vector_settings_b = await VectorSettingsFactory.create(name="vector-b", dimensions=2, dataset=dataset)

        await VectorFactory.create(value=[1.0, 2.0, 3.0], vector_settings=vector_settings_a, record=record_a)
        await VectorFactory.create(value=[4.0, 5.0], vector_settings=vector_settings_b, record=record_a)
        await VectorFactory.create(value=[1.0, 2.0], vector_settings=vector_settings_b, record=record_b)

        response = await async_client.get(
            f"/api/v1/me/datasets/{dataset.id}/records",
            params={"include": RecordInclude.vectors.value},
            headers=owner_auth_header,
        )

        assert response.status_code == 200
        assert response.json() == {
            "items": [
                {
                    "id": str(record_a.id),
                    "fields": {"text": "This is a text", "sentiment": "neutral"},
                    "metadata": None,
                    "external_id": record_a.external_id,
                    "vectors": {
                        "vector-a": [1.0, 2.0, 3.0],
                        "vector-b": [4.0, 5.0],
                    },
                    "inserted_at": record_a.inserted_at.isoformat(),
                    "updated_at": record_a.updated_at.isoformat(),
                },
                {
                    "id": str(record_b.id),
                    "fields": {"text": "This is a text", "sentiment": "neutral"},
                    "metadata": None,
                    "external_id": record_b.external_id,
                    "vectors": {
                        "vector-b": [1.0, 2.0],
                    },
                    "inserted_at": record_b.inserted_at.isoformat(),
                    "updated_at": record_b.updated_at.isoformat(),
                },
                {
                    "id": str(record_c.id),
                    "fields": {"text": "This is a text", "sentiment": "neutral"},
                    "metadata": None,
                    "external_id": record_c.external_id,
                    "vectors": {},
                    "inserted_at": record_c.inserted_at.isoformat(),
                    "updated_at": record_c.updated_at.isoformat(),
                },
            ],
            "total": 3,
        }

    @pytest.mark.skip(reason="Factory integration with search engine")
    async def test_list_current_user_dataset_records_with_include_specific_vectors(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()
        record_a = await RecordFactory.create(dataset=dataset)
        record_b = await RecordFactory.create(dataset=dataset)
        record_c = await RecordFactory.create(dataset=dataset)
        vector_settings_a = await VectorSettingsFactory.create(name="vector-a", dimensions=3, dataset=dataset)
        vector_settings_b = await VectorSettingsFactory.create(name="vector-b", dimensions=2, dataset=dataset)
        vector_settings_c = await VectorSettingsFactory.create(name="vector-c", dimensions=4, dataset=dataset)

        await VectorFactory.create(value=[1.0, 2.0, 3.0], vector_settings=vector_settings_a, record=record_a)
        await VectorFactory.create(value=[4.0, 5.0], vector_settings=vector_settings_b, record=record_a)
        await VectorFactory.create(value=[6.0, 7.0, 8.0, 9.0], vector_settings=vector_settings_c, record=record_a)
        await VectorFactory.create(value=[1.0, 2.0], vector_settings=vector_settings_b, record=record_b)
        await VectorFactory.create(value=[10.0, 11.0, 12.0, 13.0], vector_settings=vector_settings_c, record=record_b)
        await VectorFactory.create(value=[14.0, 15.0, 16.0, 17.0], vector_settings=vector_settings_c, record=record_c)

        response = await async_client.get(
            f"/api/v1/me/datasets/{dataset.id}/records",
            params={"include": f"{RecordInclude.vectors.value}:{vector_settings_a.name},{vector_settings_b.name}"},
            headers=owner_auth_header,
        )

        assert response.status_code == 200
        assert response.json() == {
            "items": [
                {
                    "id": str(record_a.id),
                    "fields": {"text": "This is a text", "sentiment": "neutral"},
                    "metadata": None,
                    "external_id": record_a.external_id,
                    "vectors": {
                        "vector-a": [1.0, 2.0, 3.0],
                        "vector-b": [4.0, 5.0],
                    },
                    "inserted_at": record_a.inserted_at.isoformat(),
                    "updated_at": record_a.updated_at.isoformat(),
                },
                {
                    "id": str(record_b.id),
                    "fields": {"text": "This is a text", "sentiment": "neutral"},
                    "metadata": None,
                    "external_id": record_b.external_id,
                    "vectors": {
                        "vector-b": [1.0, 2.0],
                    },
                    "inserted_at": record_b.inserted_at.isoformat(),
                    "updated_at": record_b.updated_at.isoformat(),
                },
                {
                    "id": str(record_c.id),
                    "fields": {"text": "This is a text", "sentiment": "neutral"},
                    "metadata": None,
                    "external_id": record_c.external_id,
                    "vectors": {},
                    "inserted_at": record_c.inserted_at.isoformat(),
                    "updated_at": record_c.updated_at.isoformat(),
                },
            ],
            "total": 3,
        }

    @pytest.mark.skip(reason="Factory integration with search engine")
    async def test_list_current_user_dataset_records_with_offset(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()
        await RecordFactory.create(fields={"record_a": "value_a"}, dataset=dataset)
        await RecordFactory.create(fields={"record_b": "value_b"}, dataset=dataset)
        record_c = await RecordFactory.create(fields={"record_c": "value_c"}, dataset=dataset)

        other_dataset = await DatasetFactory.create()
        await RecordFactory.create_batch(size=2, dataset=other_dataset)

        response = await async_client.get(
            f"/api/v1/me/datasets/{dataset.id}/records", headers=owner_auth_header, params={"offset": 2}
        )

        assert response.status_code == 200

        response_body = response.json()
        assert [item["id"] for item in response_body["items"]] == [str(record_c.id)]

    @pytest.mark.skip(reason="Factory integration with search engine")
    async def test_list_current_user_dataset_records_with_limit(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()
        record_a = await RecordFactory.create(fields={"record_a": "value_a"}, dataset=dataset)
        await RecordFactory.create(fields={"record_b": "value_b"}, dataset=dataset)
        await RecordFactory.create(fields={"record_c": "value_c"}, dataset=dataset)

        other_dataset = await DatasetFactory.create()
        await RecordFactory.create_batch(size=2, dataset=other_dataset)

        response = await async_client.get(
            f"/api/v1/me/datasets/{dataset.id}/records", headers=owner_auth_header, params={"limit": 1}
        )

        assert response.status_code == 200

        response_body = response.json()
        assert [item["id"] for item in response_body["items"]] == [str(record_a.id)]

    @pytest.mark.skip(reason="Factory integration with search engine")
    async def test_list_current_user_dataset_records_with_offset_and_limit(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()
        await RecordFactory.create(fields={"record_a": "value_a"}, dataset=dataset)
        record_c = await RecordFactory.create(fields={"record_b": "value_b"}, dataset=dataset)
        await RecordFactory.create(fields={"record_c": "value_c"}, dataset=dataset)

        other_dataset = await DatasetFactory.create()
        await RecordFactory.create_batch(size=2, dataset=other_dataset)

        response = await async_client.get(
            f"/api/v1/me/datasets/{dataset.id}/records", headers=owner_auth_header, params={"offset": 1, "limit": 1}
        )

        assert response.status_code == 200

        response_body = response.json()
        assert [item["id"] for item in response_body["items"]] == [str(record_c.id)]

    @pytest.mark.parametrize(
        ("property_config", "param_value", "expected_filter_class", "expected_filter_args"),
        [
            (
                {"name": "terms_prop", "settings": {"type": "terms"}},
                "value",
                TermsMetadataFilter,
                dict(values=["value"]),
            ),
            (
                {"name": "terms_prop", "settings": {"type": "terms"}},
                "value1,value2",
                TermsMetadataFilter,
                dict(values=["value1", "value2"]),
            ),
            (
                {"name": "integer_prop", "settings": {"type": "integer"}},
                '{"ge": 10, "le": 20}',
                IntegerMetadataFilter,
                dict(ge=10, le=20),
            ),
            (
                {"name": "integer_prop", "settings": {"type": "integer"}},
                '{"ge": 20}',
                IntegerMetadataFilter,
                dict(ge=20, le=None),
            ),
            (
                {"name": "integer_prop", "settings": {"type": "integer"}},
                '{"le": 20}',
                IntegerMetadataFilter,
                dict(ge=None, le=20),
            ),
            (
                {"name": "float_prop", "settings": {"type": "float"}},
                '{"ge": -1.30, "le": 23.23}',
                FloatMetadataFilter,
                dict(ge=-1.30, le=23.23),
            ),
            (
                {"name": "float_prop", "settings": {"type": "float"}},
                '{"ge": 23.23}',
                FloatMetadataFilter,
                dict(ge=23.23, le=None),
            ),
            (
                {"name": "float_prop", "settings": {"type": "float"}},
                '{"le": 11.32}',
                FloatMetadataFilter,
                dict(ge=None, le=11.32),
            ),
        ],
    )
    async def test_list_current_user_dataset_records_with_metadata_filter(
        self,
        async_client: "AsyncClient",
        mock_search_engine: SearchEngine,
        owner: User,
        owner_auth_header: dict,
        property_config: dict,
        param_value: str,
        expected_filter_class: Type[MetadataFilter],
        expected_filter_args: dict,
    ):
        workspace = await WorkspaceFactory.create()
        dataset, _, records, _, _ = await self.create_dataset_with_user_responses(owner, workspace)

        metadata_property = await MetadataPropertyFactory.create(
            name=property_config["name"],
            settings=property_config["settings"],
            dataset=dataset,
        )

        mock_search_engine.search.return_value = SearchResponses(
            total=2,
            items=[
                SearchResponseItem(record_id=records[0].id, score=14.2),
                SearchResponseItem(record_id=records[1].id, score=12.2),
            ],
        )

        query_params = {"metadata": [f"{metadata_property.name}:{param_value}"]}
        response = await async_client.get(
            f"/api/v1/me/datasets/{dataset.id}/records",
            params=query_params,
            headers=owner_auth_header,
        )
        assert response.status_code == 200

        response_json = response.json()
        assert response_json["total"] == 2

        mock_search_engine.search.assert_called_once_with(
            dataset=dataset,
            query=None,
            metadata_filters=[expected_filter_class(metadata_property=metadata_property, **expected_filter_args)],
            user_response_status_filter=None,
            offset=0,
            limit=LIST_DATASET_RECORDS_LIMIT_DEFAULT,
            sort_by=[SortBy(field=RecordSortField.inserted_at)],
        )

    @pytest.mark.skip(reason="Factory integration with search engine")
    @pytest.mark.parametrize("response_status_filter", ["missing", "pending", "discarded", "submitted", "draft"])
    async def test_list_current_user_dataset_records_with_response_status_filter(
        self, async_client: "AsyncClient", owner: "User", owner_auth_header: dict, response_status_filter: str
    ):
        num_responses_per_status = 10
        response_values = {"input_ok": {"value": "yes"}, "output_ok": {"value": "yes"}}

        dataset = await DatasetFactory.create()
        # missing responses
        await RecordFactory.create_batch(size=num_responses_per_status, dataset=dataset)
        # discarded responses
        await self.create_records_with_response(num_responses_per_status, dataset, owner, ResponseStatus.discarded)
        # submitted responses
        await self.create_records_with_response(
            num_responses_per_status, dataset, owner, ResponseStatus.submitted, response_values
        )
        # drafted responses
        await self.create_records_with_response(
            num_responses_per_status, dataset, owner, ResponseStatus.draft, response_values
        )

        other_dataset = await DatasetFactory.create()
        await RecordFactory.create_batch(size=2, dataset=other_dataset)

        response = await async_client.get(
            f"/api/v1/me/datasets/{dataset.id}/records?response_status={response_status_filter}&include=responses",
            headers=owner_auth_header,
        )

        assert response.status_code == 200
        response_json = response.json()

        assert len(response_json["items"]) == num_responses_per_status

        if response_status_filter in ["missing", "pending"]:
            assert all([len(record["responses"]) == 0 for record in response_json["items"]])
        else:
            assert all(
                [record["responses"][0]["status"] == response_status_filter for record in response_json["items"]]
            )

    @pytest.mark.parametrize(
        "sorts",
        [
            [("inserted_at", None)],
            [("inserted_at", "asc")],
            [("inserted_at", "desc")],
            [("updated_at", None)],
            [("updated_at", "asc")],
            [("updated_at", "desc")],
            [("metadata.terms-metadata-property", None)],
            [("metadata.terms-metadata-property", "asc")],
            [("metadata.terms-metadata-property", "desc")],
            [("inserted_at", "asc"), ("updated_at", "desc")],
            [("inserted_at", "desc"), ("updated_at", "asc")],
            [("inserted_at", "asc"), ("metadata.terms-metadata-property", "desc")],
            [("inserted_at", "desc"), ("metadata.terms-metadata-property", "asc")],
            [("updated_at", "asc"), ("metadata.terms-metadata-property", "desc")],
            [("updated_at", "desc"), ("metadata.terms-metadata-property", "asc")],
            [("inserted_at", "asc"), ("updated_at", "desc"), ("metadata.terms-metadata-property", "asc")],
            [("inserted_at", "desc"), ("updated_at", "asc"), ("metadata.terms-metadata-property", "desc")],
            [("inserted_at", "asc"), ("updated_at", "asc"), ("metadata.terms-metadata-property", "desc")],
        ],
    )
    async def test_list_current_user_dataset_records_with_sort_by(
        self,
        async_client: "AsyncClient",
        mock_search_engine: SearchEngine,
        owner: "User",
        owner_auth_header: dict,
        sorts: List[Tuple[str, Union[str, None]]],
    ):
        workspace = await WorkspaceFactory.create()
        dataset, _, records, _, _ = await self.create_dataset_with_user_responses(owner, workspace)

        expected_sorts_by = []
        for field, order in sorts:
            if field not in ("inserted_at", "updated_at"):
                field = await TermsMetadataPropertyFactory.create(name=field.split(".")[-1], dataset=dataset)
            expected_sorts_by.append(SortBy(field=field, order=order or "asc"))

        mock_search_engine.search.return_value = SearchResponses(
            total=2,
            items=[
                SearchResponseItem(record_id=records[0].id, score=14.2),
                SearchResponseItem(record_id=records[1].id, score=12.2),
            ],
        )

        query_params = {
            "sort_by": [f"{field}:{order}" if order is not None else f"{field}:asc" for field, order in sorts]
        }

        response = await async_client.get(
            f"/api/v1/me/datasets/{dataset.id}/records",
            params=query_params,
            headers=owner_auth_header,
        )
        assert response.status_code == 200
        assert response.json()["total"] == 2

        mock_search_engine.search.assert_called_once_with(
            dataset=dataset,
            query=None,
            metadata_filters=[],
            user_response_status_filter=None,
            offset=0,
            limit=LIST_DATASET_RECORDS_LIMIT_DEFAULT,
            sort_by=expected_sorts_by,
        )

    async def test_list_current_user_dataset_records_with_sort_by_with_wrong_sort_order_value(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()

        response = await async_client.get(
            f"/api/v1/me/datasets/{dataset.id}/records",
            params={"sort_by": "inserted_at:wrong"},
            headers=owner_auth_header,
        )
        assert response.status_code == 422
        assert response.json() == {
            "detail": "Provided sort order in 'sort_by' query param 'wrong' for field 'inserted_at' is not valid."
        }

    async def test_list_current_user_dataset_records_with_sort_by_with_non_existent_metadata_property(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        dataset = await DatasetFactory.create()

        response = await async_client.get(
            f"/api/v1/me/datasets/{dataset.id}/records",
            params={"sort_by": "metadata.i-do-not-exist:asc"},
            headers=owner_auth_header,
        )
        assert response.status_code == 422
        assert response.json() == {
            "detail": f"Provided metadata property in 'sort_by' query param 'i-do-not-exist' not found in dataset with '{dataset.id}'."
        }

    async def test_list_current_user_dataset_records_with_sort_by_with_invalid_field(
        self, async_client: "AsyncClient", owner: "User", owner_auth_header: dict
    ):
        workspace = await WorkspaceFactory.create()
        dataset, _, _, _, _ = await self.create_dataset_with_user_responses(owner, workspace)

        response = await async_client.get(
            f"/api/v1/me/datasets/{dataset.id}/records",
            params={"sort_by": "not-valid"},
            headers=owner_auth_header,
        )
        assert response.status_code == 422
        assert response.json() == {
            "detail": "Provided sort field in 'sort_by' query param 'not-valid' is not valid. It must be either 'inserted_at', 'updated_at' or `metadata.metadata-property-name`"
        }

    async def test_list_current_user_dataset_records_without_authentication(self, async_client: "AsyncClient"):
        dataset = await DatasetFactory.create()

        response = await async_client.get(f"/api/v1/me/datasets/{dataset.id}/records")

        assert response.status_code == 401

    @pytest.mark.skip(reason="Factory integration with search engine")
    @pytest.mark.parametrize("role", [UserRole.admin, UserRole.annotator])
    async def test_list_current_user_dataset_records_as_restricted_user(
        self, async_client: "AsyncClient", role: UserRole
    ):
        workspace = await WorkspaceFactory.create()
        user = await UserFactory.create(workspaces=[workspace], role=role)
        dataset = await DatasetFactory.create(workspace=workspace)
        record_a = await RecordFactory.create(fields={"record_a": "value_a"}, dataset=dataset)
        record_b = await RecordFactory.create(
            fields={"record_b": "value_b"}, metadata_={"unit": "test"}, dataset=dataset
        )
        record_c = await RecordFactory.create(fields={"record_c": "value_c"}, dataset=dataset)
        expected_records = [record_a, record_b, record_c]

        other_dataset = await DatasetFactory.create()
        await RecordFactory.create_batch(size=2, dataset=other_dataset)

        response = await async_client.get(
            f"/api/v1/me/datasets/{dataset.id}/records", headers={API_KEY_HEADER_NAME: user.api_key}
        )

        assert response.status_code == 200

        response_items = response.json()["items"]

        for expected_record in expected_records:
            found_items = [item for item in response_items if item["id"] == str(expected_record.id)]
            assert found_items, expected_record

            assert found_items[0] == {
                "id": str(expected_record.id),
                "fields": expected_record.fields,
                "metadata": expected_record.metadata_,
                "external_id": expected_record.external_id,
                "inserted_at": expected_record.inserted_at.isoformat(),
                "updated_at": expected_record.updated_at.isoformat(),
            }

    @pytest.mark.parametrize("role", [UserRole.annotator, UserRole.admin])
    async def test_list_current_user_dataset_records_as_restricted_user_from_different_workspace(
        self, async_client: "AsyncClient", role: UserRole
    ):
        dataset = await DatasetFactory.create()
        workspace = await WorkspaceFactory.create()
        user = await UserFactory.create(workspaces=[workspace], role=role)

        response = await async_client.get(
            f"/api/v1/me/datasets/{dataset.id}/records", headers={API_KEY_HEADER_NAME: user.api_key}
        )

        assert response.status_code == 403

    async def test_list_current_user_dataset_records_with_nonexistent_dataset_id(
        self, async_client: "AsyncClient", owner_auth_header: dict
    ):
        await DatasetFactory.create()

        response = await async_client.get(f"/api/v1/me/datasets/{uuid4()}/records", headers=owner_auth_header)

        assert response.status_code == 404



#  Copyright 2021-present, the Recognai S.L. team.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import pytest
from argilla_server.models import User
from httpx import AsyncClient

from tests.factories import UserFactory


@pytest.mark.asyncio
class TestsAuthentication:
    async def authenticate(self, async_client: AsyncClient):
        user = await UserFactory.create()

        response = await async_client.post(
            "/api/security/token",
            data={"username": user.username, "password": "12345678"},
        )
        assert response.status_code == 200
        assert response.json()["access_token"]
        assert response.json()["token_type"] == "bearer"

    async def test_invalid_credentials(self, async_client: AsyncClient, owner: User):
        response = await async_client.post(
            "/api/security/token", data={"username": owner.username, "password": "invalid"}
        )
        assert response.status_code == 401
